"""MLflow and Tensorboard utils.
"""
from pathlib import Path
from typing import Optional

import mlflow
import pandas as pd
from tensorboard.backend.event_processing import event_accumulator

SCALAR_TYPE = dict[str, list[event_accumulator.ScalarEvent]]


def read_tensorboard_scalars(events: str) -> SCALAR_TYPE:
    """Read Tensorboard scalars.

    Args:
        events: path to tensorboard events file.

    Returns:
         a dict of scalar events.
    """
    scalars = {}
    events = str(events) if isinstance(events, Path) else events
    event_acc = event_accumulator.EventAccumulator(events)
    event_acc.Reload()
    for tag in event_acc.Tags()["scalars"]:
        scalars[tag] = event_acc.Scalars(tag)
    return scalars


def _get_scalar_values(scalars: SCALAR_TYPE, tag: str) -> Optional[list[float]]:
    """Get scalar values from a dict of scalar events.

    Args:
        scalars: a dict of scalar events.
        tag: the tag of the scalar event.

    Returns:
        a list of scalar values.
    """
    if tag in scalars:
        return [acc.value for acc in scalars[tag]]
    else:
        return None


def parse_tensorboard_scalars(scalars: SCALAR_TYPE) -> list[dict]:
    """Parse Tensorboard scalars into a format to log to mlflow.

    Args:
        scalars: a dict of scalar events.

    Returns:
        a list of dict with metrics and step value
    """
    train_accuracy = _get_scalar_values(scalars, tag="train_accuracy_epoch")
    train_ce = _get_scalar_values(scalars, tag="train_cross_entropy_epoch")
    val_accuracy = _get_scalar_values(scalars, tag="val_accuracy")
    val_ce = _get_scalar_values(scalars, tag="val_cross_entropy")
    test_accuracy = _get_scalar_values(scalars, tag="test_accuracy")
    test_ce = _get_scalar_values(scalars, tag="test_cross_entropy")
    epoch = sorted(list(set(_get_scalar_values(scalars, tag="epoch"))))
    epoch = list(map(int, epoch))

    df = pd.DataFrame(
        {
            "train_accuracy": train_accuracy,
            "train_cross_entropy": train_ce,
            "val_accuracy": val_accuracy,
            "val_cross_entropy": val_ce,
            "test_accuracy": test_accuracy,
            "test_cross_entropy": test_ce,
            "epoch": epoch,
        }
    )
    cols = df.columns[df.isnull().sum(0) == 0]
    parsed = df[cols].to_dict("records")

    return parsed


def connect_to_mlflow(
    tracking_uri: str,
    experiment_name: Optional[str] = None,
    artifact_location: Optional[str] = None,
) -> None:
    """Connect to MLflow.

    Args:
        tracking_uri: MLflow tracking uri.
        experiment_name: if given, set MLflow experiment name.
        artifact_location: if experiment given, set its default artifact location
    """
    mlflow.set_tracking_uri(tracking_uri)

    if experiment_name:
        experiments = mlflow.search_experiments()
        if experiment_name in [exp.name for exp in experiments]:
            mlflow.set_experiment(experiment_name)
        else:
            mlflow.create_experiment(experiment_name, artifact_location)
            mlflow.set_experiment(experiment_name)


def get_tensorboard_files(path: str) -> tuple[list[Path], Path, Path]:
    """Get tensorboard files from a path.

    Args:
        path: path to a tensorboard log directory

    Returns:
        a tuple of paths: a list of events, hparams.yaml and latest checkpoint
    """
    path = Path(path)

    events = list(path.glob("events.*"))
    assert len(events) > 0, "Expected at least one event file"

    hparams = path / "hparams.yaml"
    assert hparams.exists(), f"{hparams=} does not exist."

    ckpt = list(path.glob("checkpoints/*.ckpt"))
    assert len(ckpt) >= 1, f"no checkpoint found in {str(path)}"
    ckpt = ckpt[-1]

    return events, hparams, ckpt



"""Triton inference server and client utils.
"""
import shutil
from pathlib import Path
from typing import List

import torch
from google.protobuf import json_format, text_format
from onnxruntime import InferenceSession


def save_triton_config(config: dict, write_file: str) -> None:
    """Dump dict config into protobuf file

    Args:
        config: a dict triton model config
        write_file: a path to write config

    Returns:
        None
    """
    from tritonclient.grpc import model_config_pb2

    parsed = json_format.ParseDict(config, model_config_pb2.ModelConfig())
    parsed_bytes = text_format.MessageToBytes(parsed)
    with open(write_file, "wb") as f:
        f.write(parsed_bytes)


def init_model_repo(triton_repo: str, triton_name: str, version: int):
    model_version_dir = Path(triton_repo) / triton_name / str(version)
    model_version_dir.mkdir(parents=True, exist_ok=True)
    return model_version_dir


def _export_txt_file(labels: List[str], filepath: Path):
    filepath.write_text("\n".join(labels))


class TritonExporter:
    def non_implemented(self):
        raise NotImplementedError("Subclass has to implement this method")

    @property
    def input_names(self):
        return ["image"]

    @property
    def output_names(self):
        return ["output"]

    @property
    def triton_batch_size(self):
        return 16

    @property
    def dynamic_axes(self):
        return {"image": {0: "batch_size"}, "output": {0: "batch_size"}}

    @property
    def triton_labels(self) -> List[str]:
        return []

    @property
    def triton_labels_path(self) -> str:
        return "labels.txt"

    def _create_triton_config(self, *args, **kwargs) -> dict:
        return {
            "backend": "onnxruntime",
            "max_batch_size": self.triton_batch_size,
        }

    @staticmethod
    def triton_python_file(self) -> str:
        self.non_implemented()

    def generate_dummy_inputs(self, *args, **kwargs) -> tuple[torch.Tensor]:
        self.non_implemented()

    def _load_model_torch(self, *args, **kwargs) -> torch.nn.Module:
        self.non_implemented()

    def _create_triton_python_backend(self):
        self.non_implemented()

    def _generate_triton_python_names(self, *args, **kwargs):
        self.non_implemented()

    def _create_triton_ensemble_config(self, *args, **kwargs) -> dict:
        self.non_implemented()

    def export_onnx(self, model_path: str, write_path: str = "auto", **kwargs):
        """Create an ONNX model from a torch model.

        Args:
            model_path: flash training checkpoint
            write_path: where to save onnx model (*.onnx extension),
                if default 'auto' then save to the same dir as model_path
                but with *.onnx extension
            kwargs: keyword arguments to pass to generating dummy inputs
        """
        # TODO create output dir if not exist
        if write_path == "auto":
            write_path = Path(model_path).with_suffix(".onnx")
            if write_path.exists():
                raise FileExistsError(f"{write_path=!r} already exists")

        model = self._load_model_torch(model_path)
        dummy = self.generate_dummy_inputs(**kwargs)
        torch.onnx.export(
            model,
            dummy,
            str(write_path),
            input_names=self.input_names,
            output_names=self.output_names,
            dynamic_axes=self.dynamic_axes,
        )

    def export_torchscirpt(self, write_path, **kwargs):
        model = self._load_model_torch(**kwargs)
        dummy = self.generate_dummy_inputs(**kwargs)
        scripted = torch.jit.script(model, example_inputs=dummy)
        torch.jit.save(scripted, write_path)

    def export_triton(
        self,
        ckpt_path: str,
        triton_repo: str,
        triton_name: str,
        version: int = 1,
        torchscript: bool = False,
        **kwargs,
    ):
        """Create a Triton model from a torch model.

        Args:
            ckpt_path: load a trained model from flash this checkpoint
            triton_repo: triton model repository path
            triton_name: triton model name
            version: triton model version
            torchscript: if True, then export in torchscript format rather
                than ONNX
        """
        model_version_dir = init_model_repo(triton_repo, triton_name, version)

        if ckpt_path is not None:
            if torchscript:
                self.export_torchscirpt(
                    write_path=model_version_dir / "model.pt", **kwargs
                )
            else:
                self.export_onnx(
                    ckpt_path, str(model_version_dir / "model.onnx"), **kwargs
                )

        if config := self._create_triton_config(torchscript=torchscript, **kwargs):
            write_config = model_version_dir.parent / "config.pbtxt"
            save_triton_config(config, str(write_config))

        if labels := self.triton_labels:
            _export_txt_file(labels, model_version_dir.parent / self.triton_labels_path)

        print(f"Triton-onnx model has been exported to {str(model_version_dir.parent)}")

    def export_triton_ensemble(
        self, triton_repo: str, triton_name: str, version: int = 1, **kwargs
    ):
        """Create a triton ensemble model.

        Args:
            triton_repo: triton model repository path
            triton_name: triton model name
            version: triton model version
            kwargs: extra params to pass to _create_triton_ensemble_config
        """
        model_version_dir = init_model_repo(triton_repo, triton_name, version)

        if config := self._create_triton_ensemble_config(**kwargs):
            write_config = model_version_dir.parent / "config.pbtxt"
            save_triton_config(config, str(write_config))

        print(f"Triton-ensemble has been exported to {str(model_version_dir.parent)}")

    def export_triton_python(
        self, triton_repo: str, triton_name: str, version: int = 1, **kwargs
    ):
        """Create a triton model with python backend.

        Args:
            triton_repo: triton model repository path
            triton_name: triton model name
            version: triton model version
            **kwargs: extra params to pass to _generate_triton_python_names
        """
        model_version_dir = init_model_repo(triton_repo, triton_name, version)

        shutil.copy(
            self.triton_python_file,
            model_version_dir / "model.py",
        )

        if config := self._create_triton_python_backend():
            write_config = model_version_dir.parent / "config.pbtxt"
            save_triton_config(config, write_config)

        if names := self._generate_triton_python_names(**kwargs):
            assert bool(kwargs), "Provide kwargs with dependent names"
            _export_txt_file(names, model_version_dir.parent / "names.txt")

        if "labels" in kwargs:
            shutil.copy(
                kwargs["labels"],
                model_version_dir.parent / self.triton_labels_path,
            )

        print(
            f"Triton-python model has been exported to {str(model_version_dir.parent)}"
        )

    @staticmethod
    def print_onnx(model_path: str):
        """Print ONNX model inputs and outputs.

        Args:
            model_path: path to the ONNX model
        """
        ort = InferenceSession(model_path)
        print("INPUTS:")
        for inp in ort.get_inputs():
            print(">>>", inp)
        print("\nOUTPUTS:")
        for out in ort.get_outputs():
            print(">>>", out)



"""Create tables/collections in a DB.
"""
from pathlib import Path
from typing import Any, Dict, List, Union

import fire

from finegrained.utils.milvus import MilvusClient
from finegrained.utils.os_utils import read_file_config


def milvus(collection_name: str, fields: Union[List[Dict[str, Any]], str]):
    if isinstance(fields, str) and Path(fields).exists():
        fields = read_file_config(fields, section=collection_name)
    milvus = MilvusClient()
    milvus.create_collection("test_milvus", fields)


if __name__ == "__main__":
    fire.Fire({"milvus": milvus})



"""Send, query and get annotation results.
"""
from pathlib import Path
from typing import Any, Optional

from ..utils import types
from ..utils.dataset import load_fiftyone_dataset
from ..utils.os_utils import read_file_config, read_txt


def _load_backend_config(src: Any) -> dict:
    if isinstance(src, dict):
        config = src
    elif isinstance(src, str) and Path(src).is_file():
        # if a txt file as file
        config = read_file_config(src, section="annotation_backend")
    else:
        raise ValueError(f"{type(src)!r} type has not been implemented.")
    return config


def _parse_classes(src: Any) -> list[str]:
    if isinstance(src, list):
        return src
    elif isinstance(src, str) and Path(src).is_file():
        return read_txt(src)
    else:
        raise ValueError(
            "Classes has to be a list of str or a path to a txt file."
            f"Received {type(src)}"
        )


def annotate(
    dataset: str,
    annotation_key: str,
    label_field: str,
    backend: Any,
    overwrite: bool = False,
    label_type: Optional[str] = None,
    project_id: Optional[int] = None,
    segment_size: int = 10,
    task_name: Optional[str] = None,
    image_quality: int = 75,
    task_asignee: Optional[str] = None,
    classes: Optional[str] = None,
    **kwargs,
):
    """Send samples to annotations

    Args:
        dataset: fiftyone dataset with samples
        annotation_key: assign this key for annotation run
        label_field: if exists, upload labels
        label_type: if label_field does not exist, this has to be specified
        backend: backend name or filepath to configs
        overwrite: overwrite existing annotation run if True
        classes: list of classes or path to labels.txt file
        image_quality: image upload quality
        task_name: custom task name, by default dataset name + annotation key
        segment_size: number of frames/images per one job
        project_id: which cvat project to connect to
        task_asignee: assignee for the task
        **kwargs: dataset loading filters
    """
    dataset = load_fiftyone_dataset(dataset, **kwargs)
    backend_conf = _load_backend_config(backend)
    if dataset.has_annotation_run(annotation_key) and overwrite:
        dataset.delete_annotation_run(annotation_key)
    if not dataset.has_sample_field(label_field) and label_type is None:
        raise ValueError(
            f"{label_field=} does not exist in {dataset.name}. Specify 'label_type'"
        )
    if classes:
        classes = _parse_classes(classes)
    if task_name is None:
        task_name = dataset.name + " - " + annotation_key
    dataset.annotate(
        annotation_key,
        label_field=label_field,
        label_type=label_type,
        project_id=project_id,
        segment_size=segment_size,
        task_name=task_name,
        image_quality=image_quality,
        classes=classes,
        task_asignee=task_asignee,
        **backend_conf,
    )


def load(
    dataset: str,
    annotation_key: str,
    backend: Any,
    dest_field: str = None,
    dataset_kwargs: Optional[dict] = None,
):
    """Download annotations from an annotation backend.

    Args:
        dataset: fiftyone dataset name
        annotation_key: annotation key used to send for annotations
        backend: annotation backend name or filepath with configs
        dest_field: if given, annotations will be stored in a new field
        dataset_kwargs: dataset loading filters

    Returns:
        none
    """
    backend_conf = _load_backend_config(backend)
    backend_conf.pop("backend")
    dataset = load_fiftyone_dataset(
        dataset, **dataset_kwargs if bool(dataset_kwargs) else {}
    )
    dataset.load_annotations(annotation_key, dest_field=dest_field, **backend_conf)


def list_keys(dataset: str) -> types.LIST_STR:
    """List annotation keys attributed to the dataset

    Args:
        dataset: fiftyone dataset name

    Returns:
        a list of keys
    """
    dataset = load_fiftyone_dataset(dataset)
    keys = dataset.list_annotation_runs()
    return keys


def delete_key(dataset: str, key: str):
    """Delete an annotation key.

    Args:
        dataset: fiftyone dataset name
        key: annotation key

    Returns:
        none
    """
    dataset = load_fiftyone_dataset(dataset)
    dataset.delete_annotation_run(key)


def print_status(dataset: str, key: str, backend: str) -> None:
    dataset = load_fiftyone_dataset(dataset)
    assert dataset.has_annotation_run(key)

    backend_conf = _load_backend_config(backend)
    backend_conf.pop("backend")

    results = dataset.load_annotation_results(key, **backend_conf)
    results.print_status()



"""Main module."""
import fire

from finegrained import data, models, services


def main():
    fire.Fire({"data": data, "models": models, "services": services})


if __name__ == "__main__":
    main()



"""Train, evaluate and predict with image meta-learning .
"""
from typing import Tuple

import fiftyone.core.labels as fol
import torch
from fiftyone import Dataset
from flash import DataModule
from flash.image import ImageClassificationData, ImageClassifier

from finegrained.models.flash_base import FlashFiftyOneTask
from finegrained.models.flash_transforms import get_transform
from finegrained.utils import types
from finegrained.utils.dataset import get_all_filepaths, load_fiftyone_dataset
from finegrained.utils.triton import TritonExporter

# TODO turn into a class
# TODO split before training


def _init_support_datamodule(
    dataset: str,
    label_field: str,
    image_size: Tuple[int, int],
    batch_size: int = 4,
    **kwargs,
) -> DataModule:
    dataset = load_fiftyone_dataset(dataset, **kwargs)

    data = ImageClassificationData.from_fiftyone(
        train_dataset=dataset,
        label_field=label_field,
        batch_size=batch_size,
        transform_kwargs=dict(image_size=image_size),
    )
    return data


def _init_query_datamodule(
    dataset: str, image_size: Tuple[int, int], batch_size: int = 4, **kwargs
) -> Tuple[DataModule, Dataset]:
    dataset = load_fiftyone_dataset(dataset, **kwargs)
    data = ImageClassificationData.from_fiftyone(
        predict_dataset=dataset,
        batch_size=batch_size,
        transform_kwargs=dict(image_size=image_size),
    )
    return data, dataset


class ImageMetalearn(FlashFiftyOneTask, TritonExporter):
    """Image meta learning."""

    @property
    def model_keys(self):
        return ["backbone", "training_strategy", "training_strategy_kwargs"]

    def _init_training_datamodule(self, dataset: str, label_field: str, **kwargs):
        dataset = load_fiftyone_dataset(dataset)
        self.labels = dataset.match_tags(["train_metalearn", "val_metalearn"]).distinct(
            f"{label_field}.label"
        )

        def get_targets(tag):
            targets = [
                self.labels.index(smp[label_field].label)
                for smp in dataset.match_tags(tag).select_fields(label_field)
            ]
            return targets

        train_files = get_all_filepaths(dataset.match_tags("train_metalearn"))
        train_targets = get_targets("train_metalearn")
        val_files = get_all_filepaths(dataset.match_tags("val_metalearn"))
        val_targets = get_targets("val_metalearn")

        # use from_files because of label mapping error with from_fiftyone
        self.data = ImageClassificationData.from_files(
            train_files=train_files,
            train_targets=train_targets,
            val_files=val_files,
            val_targets=val_targets,
            batch_size=kwargs.get("batch_size", 16),
            transform=get_transform(kwargs.get("transform")),
            transform_kwargs=kwargs.get("transform_kwargs"),
        )

    def _init_prediction_datamodule(
        self,
        support_dataset: str,
        support_label: str,
        query_dataset: str,
        image_size: types.IMAGE_SIZE,
        batch_size: int,
        support_kwargs={},
        query_kwargs={},
    ):
        self.support_data = _init_support_datamodule(
            support_dataset,
            support_label,
            image_size=image_size,
            batch_size=batch_size,
            **support_kwargs,
        )
        self.query_data, self.prediction_dataset = _init_query_datamodule(
            query_dataset,
            image_size=image_size,
            batch_size=batch_size,
            **query_kwargs,
        )

    def _init_model(
        self,
        backbone: str,
        training_strategy: str,
        training_strategy_kwargs: dict,
        **kwargs,
    ):
        self.model = ImageClassifier(
            num_classes=len(self.labels),
            backbone=backbone,
            labels=self.labels,
            training_strategy=training_strategy,
            training_strategy_kwargs=training_strategy_kwargs,
            **kwargs,
        )

    def _load_pretrained_model(self, ckpt_path: str):
        self.model = ImageClassifier.load_from_checkpoint(ckpt_path)

    def _predict(self):
        from learn2learn.nn import PrototypicalClassifier

        feature_extractor = self.model.backbone
        support_features, support_labels = self.calculate_features(
            feature_extractor, self.support_data.train_dataloader()
        )
        query_features, _ = self.calculate_features(
            feature_extractor, self.query_data.predict_dataloader()
        )

        clf = PrototypicalClassifier(support=support_features, labels=support_labels)
        dist = clf(query_features)
        class_probs = dist.softmax(dim=1)
        conf, index = class_probs.max(dim=1)

        def _get_class_label(index):
            return self.support_data.train_dataset.labels[index]

        predictions = [
            fol.Classification(
                label=_get_class_label(i.item()),
                confidence=c.item(),
                logits=l.cpu().numpy(),
            )
            for c, i, l in zip(conf, index, dist)
        ]

        return predictions

    def predict(
        self,
        support_dataset: str,
        support_label_field: str,
        query_dataset: str,
        query_label_field: str,
        ckpt_path: str,
        image_size: types.IMAGE_SIZE = (224, 224),
        batch_size: int = 4,
        support_kwargs={},
        query_kwargs={},
    ):
        """Classify samples from a dataset and assign values to label field

        Args:
            support_dataset: support fiftyone dataset
            support_label_field: label field for the support dataset
            query_dataset: query fiftyone dataset
            query_label_field: where to assign predictions on the query dataset
            ckpt_path: flash model checkpoint path
            image_size: Image size for inference
            batch_size: predictions batch size
            support_kwargs: support dataset filters
            query_kwargs: query dataset filters
        """
        self._init_prediction_datamodule(
            support_dataset=support_dataset,
            support_label=support_label_field,
            query_dataset=query_dataset,
            image_size=image_size,
            batch_size=batch_size,
            support_kwargs=support_kwargs,
            query_kwargs=query_kwargs,
        )
        self._load_pretrained_model(ckpt_path)
        predictions = self._predict()
        self.prediction_dataset.set_values(query_label_field, predictions)
        print(
            f"{len(predictions)} predictions saved to "
            f"{query_label_field} field in {query_dataset}"
        )

    def _get_available_backbones(self):
        return ImageClassifier.available_backbones()

    def _load_model_torch(self, ckpt_path) -> torch.nn.Module:
        self._load_pretrained_model(ckpt_path)
        return self.model.backbone

    def generate_dummy_inputs(
        self, image_size: int
    ) -> tuple[torch.Tensor,]:
        return (torch.rand(1, 3, image_size, image_size),)

    def _create_triton_ensemble_config(
        self, preprocessing_name: str, feature_extractor: str, embedding_size: int
    ) -> dict:
        return {
            "platform": "ensemble",
            "max_batch_size": 1,
            "input": [dict(name="IMAGE", dims=[-1, -1, 3], data_type="TYPE_UINT8")],
            "output": [
                dict(
                    name="FEATURES",
                    dims=[embedding_size],
                    data_type="TYPE_FP32",
                )
            ],
            "ensemble_scheduling": dict(
                step=[
                    dict(
                        model_name=preprocessing_name,
                        model_version=-1,
                        input_map=dict(image="IMAGE"),
                        output_map=dict(output="PREPROCESSED"),
                    ),
                    dict(
                        model_name=feature_extractor,
                        model_version=-1,
                        input_map=dict(image="PREPROCESSED"),
                        output_map=dict(output="FEATURES"),
                    ),
                ]
            ),
        }



"""Utilities to work with Milvus vector search.
"""
from typing import Any, Dict, List

import numpy as np
from pymilvus import (
    Collection,
    CollectionSchema,
    DataType,
    FieldSchema,
    connections,
    utility,
)


class MilvusClient:
    """A class to connect to Milvus and perform CRUD ops."""

    def __init__(self, **kwargs):
        connections.connect(**kwargs)
        self.collections = {}

    def get_collection(self, collection_name: str):
        if collection_name not in self.collections:
            raise ValueError(f"Collection {collection_name!r} does not exist.")
        return self.collections[collection_name]

    @staticmethod
    def _convert_dtype(field: Dict[str, Any]):
        if "dtype" in field:
            field.update({"dtype": _to_milvus_datatype(field["dtype"])})

    def create_collection(
        self, name: str, fields: List[Dict[str, Any]], description=None
    ):
        """Create a new collection."""
        _ = list(map(self._convert_dtype, fields))
        fields = [FieldSchema(**field) for field in fields]
        schema = CollectionSchema(fields=fields, description=description)
        collection = Collection(name, schema)
        self.collections.update({name: collection})

    @staticmethod
    def drop_collection(name: str):
        utility.drop_collection(name)

    def build_index(self, field: str, collection_name: str, **kwargs):
        collection = self.get_collection(collection_name)
        collection.create_index(field, kwargs)

    def insert_vectors(self, data: List[List[Any]], collection_name: str):
        collection = self.get_collection(collection_name)
        collection.insert(data)

    def search_vectors(
        self,
        vectors: List[List[float]],
        collection_name: str,
        metric_type="l2",
        top_k=5,
        **kwargs,
    ):
        collection = self.get_collection(collection_name)
        collection.load()
        search_params = {
            "metric_type": metric_type,
            "params": {"nprobe": top_k},
        }
        result = collection.search(vectors, search_params, **kwargs)
        return result


def _to_milvus_datatype(data_type):
    if data_type in [str, np.object, np.str]:
        return DataType.NONE
    elif data_type in [bool]:
        return DataType.BOOL
    elif data_type in [int, np.int64]:
        return DataType.INT64
    elif data_type in [np.int32]:
        return DataType.INT32
    elif data_type in [float, np.float64]:
        return DataType.DOUBLE
    elif data_type in [np.float32]:
        return DataType.FLOAT_VECTOR
    else:
        raise ValueError(f"Unsupported data type {data_type}")



"""Tag or untag samples with specific filters or condition
"""
from pathlib import Path
from typing import Optional

from fiftyone import ViewField as F
from fiftyone.utils import random as four
from sklearn.model_selection import train_test_split

from finegrained.data.display import label_diff
from finegrained.utils import types
from finegrained.utils.dataset import load_fiftyone_dataset
from finegrained.utils.general import parse_list_str


def tag_samples(dataset: str, tags: types.LIST_STR_STR, **kwargs) -> dict:
    """Tag each sample in dataset with given tags

    Args:
        dataset: fiftyone dataset name
        tags: tags to apply
        kwargs: dataset loading kwargs, i.e. filters

    Returns:
        a dict of sample tag counts
    """
    dataset = load_fiftyone_dataset(dataset, **kwargs)
    dataset.tag_samples(parse_list_str(tags))
    return dataset.count_sample_tags()


def split_dataset(
    dataset: str,
    splits: types.DICT_STR_FLOAT = {"train": 0.8, "val": 0.1, "test": 0.1},
    **kwargs,
):
    """Create data split tags for a dataset

    Args:
        dataset: fiftyone dataset
        splits: a dict of split names and relative sizes
        kwargs: dataset loading filters

    Returns:
        a dict of split counts
    """
    dataset = load_fiftyone_dataset(dataset, **kwargs)
    four.random_split(dataset, splits)
    return dataset.count_sample_tags()


def split_classes(
    dataset: str,
    label_field: str,
    train_size: float = 0.5,
    val_size: float = 0.5,
    min_samples: int = 3,
    split_names: tuple[str, str] = ("train", "val"),
    overwrite: bool = False,
) -> types.DICT_STR_FLOAT:
    """Split classes in a dataset into train and val.

    Used for meta-learning.

    Args:
        dataset: fiftyone dataset name
        label_field: which field to use for classes
        train_size: fraction of classes to tag as train
        val_size: fraction of classes to tag as val
        min_samples: minimum number of samples
            per class to include a class into a split
        split_names: splits will be tagged with these names
        overwrite: if True, existing tags are removed

    Returns:
        a dict of tag counts
    """
    dataset = load_fiftyone_dataset(dataset)
    label_counts = dataset.count_values(f"{label_field}.label")
    labels = list(filter(lambda x: label_counts[x] >= min_samples, label_counts))
    train_labels, val_labels = train_test_split(
        labels, test_size=val_size, train_size=train_size, shuffle=True
    )
    if overwrite:
        dataset.untag_samples(split_names)
    train_view = dataset.filter_labels(label_field, F("label").is_in(train_labels))
    train_view.tag_samples(split_names[0])
    val_view = dataset.filter_labels(label_field, F("label").is_in(val_labels))
    val_view.tag_samples(split_names[1])
    return dataset.count_sample_tags()


def tag_alignment(
    dataset: str, vertical: bool = True, tag: Optional[str] = None, **kwargs
) -> dict:
    """Add a vertical/horizontal tag each sample.

    Args:
        dataset: fiftyone dataset name
        vertical: if True, vertical images are tagged.
            If False, horizontal images are tagged.
        tag: overwrite default 'vertical' or 'horizontal' tag.
        **kwargs: dataset filter kwargs

    Returns:
        a dict with sample tag counts
    """
    dataset = load_fiftyone_dataset(dataset, **kwargs)
    dataset.compute_metadata()
    if vertical:
        tag = "vertical" if tag is None else tag
        tag_view = dataset.match(F("metadata.height") > F("metadata.width"))
    else:
        tag = "horizontal" if tag is None else tag
        tag_view = dataset.match(F("metadata.width") >= F("metadata.height"))
    tag_view.tag_samples(tag)
    return tag_view.count_sample_tags()


def retag_missing_labels(
    dataset: str,
    label_field: str,
    from_tags: types.LIST_STR_STR,
    to_tags: types.LIST_STR_STR,
) -> dict:
    """Remove from_tags and add to_tags for labels that are present in
        from_tags but absent in to_tags.

    Args:
        dataset: fiftyone dataset name
        label_field: a label field
        from_tags: tags with base list of class labels
        to_tags: tags with intersection of class labels

    Returns:
        a count of sample tags for a subset
    """
    # TODO test this
    diff = label_diff(dataset, label_field, tags_left=from_tags, tags_right=to_tags)
    assert len(diff) > 0, "No samples to retag"

    dataset = load_fiftyone_dataset(dataset, include_labels={label_field: diff})
    dataset.untag_samples(from_tags)
    dataset.tag_samples(to_tags)

    return dataset.count_sample_tags()


def tag_labels(
    dataset: str,
    label_field: str,
    labels: types.LIST_STR_STR,
    tags: types.LIST_STR_STR,
) -> dict:
    """Tag labels with given tags.

    Args:
        dataset: fiftyone dataset name
        label_field: a label field
        labels: labels to filter, can be a txt file with labels
        tags: tags to apply

    Returns:
        a count of label tags for a subset
    """
    if (lab := Path(labels)).is_file():
        labels = lab.read_text().strip().split("\n")
    dataset = load_fiftyone_dataset(dataset, include_labels={label_field: labels})
    dataset.tag_labels(tags, label_fields=label_field)
    return dataset.count_label_tags(label_fields=label_field)



import json
from configparser import ConfigParser
from pathlib import Path
from typing import Optional, Union

import yaml

# TODO rename to io.py


def read_txt(filepath: str):
    return Path(filepath).read_text().strip().split("\n")


def parse_keyval_line(line: str) -> tuple:
    assert line.count("=") == 1, f"{line=} has to contain exactly on '='"
    key, val = line.split("=")
    return key, val


def read_file_config(file: str, section=None):
    if file.endswith((".txt", ".text")):
        config = read_txt_credentials(file)
    elif file.endswith(".ini"):
        config = read_ini_credentials(file, section=section)
    else:
        raise NotImplementedError(f"{file=} file extension not supported yet.")
    return config


def read_txt_credentials(path: str) -> dict:
    lines = read_txt(path)
    config = list(map(parse_keyval_line, lines))
    config = dict(config)
    return config


def read_ini_credentials(
    path: str, section: Optional[str] = None
) -> Union[dict, ConfigParser]:
    parser = ConfigParser()
    parser.read(path)
    if section:
        assert section in parser.sections()
        config = {k: v for k, v in parser[section].items()}
        return config
    else:
        return parser


def read_yaml(path: str | Path):
    with open(path, "r") as f:
        data = yaml.safe_load(f)
    return data


def write_yaml(data: dict, path: str):
    with open(path, "w") as f:
        yaml.safe_dump(data, f)


def read_json(file: str) -> list | dict:
    with open(file) as f:
        data = json.load(f)
    return data


def write_json(data: list | dict, file: str):
    with open(file, "w") as f:
        json.dump(data, f)



"""Train, evaluate and predict with image classification models.
"""
import tempfile
from pathlib import Path
from typing import Tuple

import cv2
import fiftyone as fo
import fiftyone.utils.patches as foup
import numpy as np
import torch
from flash.core.data.utilities.classification import SingleLabelTargetFormatter
from flash.image import ImageClassificationData, ImageClassifier
from PIL import Image
from torchvision import transforms as T
from tqdm import tqdm

from finegrained.models.flash_base import FlashFiftyOneTask
from finegrained.models.flash_transforms import get_transform
from finegrained.utils import types
from finegrained.utils.dataset import get_unique_labels, load_fiftyone_dataset
from finegrained.utils.triton import TritonExporter


def parse_image_size(image_size: types.IMAGE_SIZE) -> Tuple[int, int]:
    """Return height and width from a tuple or int image size

    Args:
        image_size: a int size or a tuple of (height, width)

    Returns:
        a tuple of height and width
    """
    if isinstance(image_size, int):
        return image_size, image_size
    elif isinstance(image_size, (tuple, list)):
        assert (
            len(image_size) == 2
        ), f"if tuple/list should be 2 elements, got {image_size=}"
        return image_size
    else:
        raise ValueError(f"{image_size=} of type {type(image_size)} not understood")


def _get_patch_array(patch_sample, patch_field, transpose=False) -> np.ndarray:
    img = np.array(Image.open(patch_sample.filepath))
    img_h, img_w = img.shape[:2]
    x, y, w, h = patch_sample[patch_field].bounding_box
    xmin = max(0, int(img_w * x))
    ymin = max(0, int(img_h * y))
    xmax = min(img_w, int(img_w * (x + w)))
    ymax = min(img_h, int(img_h * (y + h)))
    crop = img[ymin:ymax, xmin:xmax]
    if transpose:
        crop = crop.transpose(2, 0, 1)
    return crop


def read_rgb(path: str) -> np.ndarray:
    img = cv2.imread(path)
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    return img


def _to_patches(dataset: fo.Dataset, patch_field: str, write_dir: Path):
    if not write_dir.exists():
        write_dir.mkdir(parents=True)

    i = 0
    for smp in tqdm(dataset, "extracting patches"):
        img = read_rgb(smp.filepath)
        for det in smp[patch_field].detections:
            patch = foup.extract_patch(img, det)
            path = write_dir / f"{i:04d}.jpg"
            cv2.imwrite(str(path), patch)
            det["patch_filepath"] = str(path)
            i += 1
        smp.save()


class ImageClassification(FlashFiftyOneTask, TritonExporter):
    """Image classification task."""

    def _init_training_datamodule(
        self,
        dataset: str,
        label_field: str,
        tags: types.DICT_STR_STR = {
            "train": "train",
            "val": "val",
            "test": "test",
        },
        **kwargs,
    ):
        dataset = load_fiftyone_dataset(dataset)
        dataset_tags = list(tags.values())
        self.labels = get_unique_labels(dataset.match_tags(dataset_tags), label_field)
        self.data = ImageClassificationData.from_fiftyone(
            train_dataset=dataset.match_tags(tags["train"]),
            val_dataset=dataset.match_tags(tags["val"]),
            test_dataset=dataset.match_tags(tags["test"])
            if "test" in dataset_tags
            else None,
            label_field=label_field,
            batch_size=kwargs.get("batch_size", 16),
            target_formatter=SingleLabelTargetFormatter(
                labels=self.labels, num_classes=len(self.labels)
            ),
            transform=get_transform(kwargs.get("train_transform")),
            transform_kwargs=kwargs.get("transform_kwargs"),
        )

    def _init_prediction_datamodule(
        self,
        dataset: str,
        image_size: Tuple[int, int],
        batch_size: int = 4,
        patch_field: str = None,
        **kwargs,
    ):
        self.prediction_dataset = load_fiftyone_dataset(dataset, **kwargs)
        if bool(patch_field):
            assert self.prediction_dataset.has_sample_field(patch_field)

            export_dir = Path(tempfile.TemporaryDirectory().name).resolve() / "export"

            _to_patches(self.prediction_dataset, patch_field, export_dir)
            self.patch_dataset = fo.Dataset.from_images_dir(str(export_dir))
            self.patch_dataset.persistent = False

            self.data = ImageClassificationData.from_fiftyone(
                predict_dataset=self.patch_dataset,
                batch_size=batch_size,
                transform_kwargs=dict(image_size=image_size),
            )
        else:
            self.data = ImageClassificationData.from_fiftyone(
                predict_dataset=self.prediction_dataset,
                batch_size=batch_size,
                transform_kwargs=dict(image_size=image_size),
            )

    def _init_model(self, backbone: str, **kwargs) -> ImageClassifier:
        self.model = ImageClassifier(
            num_classes=len(self.labels),
            backbone=backbone,
            labels=self.labels,
            **kwargs,
        )

    def _load_pretrained_model(self, ckpt_path: str) -> ImageClassifier:
        self.model = ImageClassifier.load_from_checkpoint(ckpt_path)

    def _get_available_backbones(self):
        return ImageClassifier.available_backbones()

    def _load_model_torch(self, ckpt_path: str) -> torch.nn.Module:
        self._load_pretrained_model(ckpt_path)
        model = SoftmaxClassifier(self.model)
        return model

    def generate_dummy_inputs(self, image_size) -> Tuple[torch.Tensor]:
        h, w = parse_image_size(image_size)
        return (torch.randn(2, 3, h, w),)

    def _create_triton_config(self, image_size: types.IMAGE_SIZE, **kwargs) -> dict:
        h, w = parse_image_size(image_size)
        return {
            "backend": "onnxruntime",
            "max_batch_size": self.triton_batch_size,
            "input": [
                dict(
                    name=self.input_names[0],
                    data_type="TYPE_FP32",
                    dims=[3, h, w],
                )
            ],
            "output": [
                dict(
                    name=self.output_names[0],
                    data_type="TYPE_FP32",
                    dims=[self.model.num_classes],
                    label_filename=self.triton_labels_path,
                )
            ],
        }

    @property
    def triton_labels(self) -> types.LIST_STR:
        labels = self.model.hparams["labels"]
        return labels

    @property
    def triton_python_file(self) -> str:
        return (
            Path(__file__).parents[1]
            / "utils"
            / "triton_python"
            / "image_classification.py"
        )

    def _create_triton_python_backend(self):
        return {
            "backend": "python",
            "max_batch_size": self.triton_batch_size,
            "input": [dict(name="IMAGE", dims=[-1, -1, 3], data_type="TYPE_UINT8")],
            "output": [
                dict(
                    name="CLASS_PROBS",
                    dims=[-1],
                    data_type="TYPE_FP32",
                    label_filename=self.triton_labels_path,
                )
            ],
        }

    def _generate_triton_python_names(
        self, preprocessing_name: str, classifier_name: str, **kwargs
    ):
        return [
            f"preprocessing={preprocessing_name}",
            f"classifier={classifier_name}",
        ]

    def _create_triton_ensemble_config(
        self, preprocessing_name: str, classifier_name: str
    ) -> dict:
        return {
            "platform": "ensemble",
            "max_batch_size": 1,
            "input": [dict(name="IMAGE", dims=[-1, -1, 3], data_type="TYPE_UINT8")],
            "output": [
                dict(
                    name="CLASS_PROBS",
                    dims=[-1],
                    data_type="TYPE_FP32",
                )
            ],
            "ensemble_scheduling": dict(
                step=[
                    dict(
                        model_name=preprocessing_name,
                        model_version=-1,
                        input_map=dict(image="IMAGE"),
                        output_map=dict(output="PREPROCESSED"),
                    ),
                    dict(
                        model_name=classifier_name,
                        model_version=-1,
                        input_map=dict(image="PREPROCESSED"),
                        output_map=dict(output="CLASS_PROBS"),
                    ),
                ]
            ),
        }


class ImageTransform(torch.nn.Module, TritonExporter):
    def __init__(self, size: int):
        super(ImageTransform, self).__init__()
        self._ = torch.nn.Sequential()
        self.transforms = T.Compose(
            [
                T.Lambda(lambda x: x.permute(2, 0, 1) / 255.0),
                T.Resize(size),
                T.RandomCrop(size),
                T.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
            ]
        )
        self.image_size = size

    def forward(self, samples: torch.Tensor) -> torch.Tensor:
        transformed = self.transforms(samples.squeeze(0))
        batch = transformed.unsqueeze(0)
        return batch

    def generate_dummy_inputs(self, **kwargs) -> tuple[torch.Tensor]:
        h, w = parse_image_size(self.image_size)
        dummy = torch.randint(
            0, 255, size=(1, h * 2 + 1, w * 3 - 10, 3), dtype=torch.uint8
        )
        return (dummy,)

    @property
    def dynamic_axes(self):
        return {"image": [0, 1, 2], "output": [0]}

    def _load_model_torch(self, *args, **kwargs) -> torch.nn.Module:
        return self

    @property
    def triton_batch_size(self):
        return 1


class SoftmaxClassifier(torch.nn.Module):
    def __init__(self, model):
        super(SoftmaxClassifier, self).__init__()
        self.model = model
        self.softmax = torch.nn.Softmax(dim=1)

    def forward(self, x):
        logits = self.model(x)
        probs = self.softmax(logits)
        return probs



"""Utils for similarity search.
"""
from copy import deepcopy
from dataclasses import dataclass
from typing import Dict, List

import numpy as np
from scipy.spatial.distance import cdist

from finegrained.models.transformers_base import SentenceEmbeddings


@dataclass
class SimilaritySearch:
    data: List[Dict]
    embeddings: np.ndarray

    def find_similar(self, query_vector: np.ndarray, top_k=5):
        query_vector = np.atleast_2d(query_vector)
        # TODO normalized distances
        dist = cdist(query_vector, self.embeddings)[0]
        top_inds = np.argsort(dist)[:top_k]
        items = [self._get_item(i, dist[i]) for i in top_inds]
        return items

    def _get_item(self, index: int, distance: float):
        d = deepcopy(self.data[index])
        d.update(dict(distance=distance))
        return d

    def embed_and_find_similar(self, query: str, model: SentenceEmbeddings, **kwargs):
        query_embed = model.embed(query).numpy()
        return self.find_similar(query_embed, **kwargs)



"""Extend Flash modules and classes.
"""
from dataclasses import dataclass
from typing import Callable, Tuple, Union

import torch
import torchvision.transforms as T
from flash.core.data.io.input_transform import InputTransform
from flash.image import ImageClassificationInputTransform


@dataclass
class RandAugmentTransform(InputTransform):
    """Applies RandAugment transformation for input images."""

    image_size: Tuple[int, int] = (224, 224)
    mean: Union[float, Tuple[float, float, float]] = (0.485, 0.456, 0.406)
    std: Union[float, Tuple[float, float, float]] = (0.229, 0.224, 0.225)
    num_ops: int = 3
    magnitude: int = 9

    def input_per_sample_transform(self):
        return T.Compose(
            [
                T.ToTensor(),
                T.Resize(self.image_size),
                T.Normalize(self.mean, self.std),
            ]
        )

    def train_input_per_sample_transform(self):
        return T.Compose(
            [
                T.RandAugment(num_ops=self.num_ops, magnitude=self.magnitude),
                T.ToTensor(),
                T.Resize(self.image_size),
                T.RandomHorizontalFlip(),
                T.Normalize(self.mean, self.std),
            ]
        )

    def target_per_sample_transform(self) -> Callable:
        return torch.as_tensor


TRANSFORMS = dict(randaugment=RandAugmentTransform)


def get_transform(name: str) -> InputTransform:
    """Get image transform definition

    If given transform exists, then return it.
    Return basic image transforms otherwise.

    Args:
        name: name of transform definition

    Returns:
        Flash InputTransform object
    """
    if name is None:
        return ImageClassificationInputTransform
    return TRANSFORMS.get(name.lower(), ImageClassificationInputTransform)



"""Transformers tokenizer in triton's python backend format
"""
from pathlib import Path

import numpy as np
import triton_python_backend_utils as pb_utils


def _read_names(repo) -> dict:
    names = (Path(repo) / "names.txt").read_text().strip().split("\n")
    names = {one.split("=")[0]: one.split("=")[1] for one in names}
    return names


# TODO unable to send batched requests with varying sizes
class TritonPythonModel:
    @staticmethod
    def auto_complete_config(auto_complete_model_config):
        inputs = [dict(name="IMAGE", dims=[-1, -1, 3], data_type="TYPE_UINT8")]
        outputs = [dict(name="CLASS_PROBS", dims=[-1], data_type="TYPE_FP32")]

        config = auto_complete_model_config.as_dict()
        existing_inputs = [inp["name"] for inp in config["input"]]
        existing_outputs = [inp["name"] for inp in config["output"]]

        for inp in inputs:
            if inp["name"] not in existing_inputs:
                auto_complete_model_config.add_input(inp)

        for out in outputs:
            if out["name"] not in existing_outputs:
                auto_complete_model_config.add_output(out)

        return auto_complete_model_config

    def initialize(self, args):
        self._names = _read_names(args["model_repository"])

    def execute(self, requests):
        responses = [self._process_request(one) for one in requests]
        return responses

    def _process_request(self, request):
        images = pb_utils.get_input_tensor_by_name(request, "IMAGE")
        images = images.as_numpy()

        preprocessed = [self._preprocess(img) for img in images]
        preprocessed = np.vstack(preprocessed)
        class_probs = self._classify(preprocessed)

        output = pb_utils.Tensor("CLASS_PROBS", class_probs)
        resp = pb_utils.InferenceResponse(output_tensors=[output])
        return resp

    def _preprocess(self, image: np.ndarray) -> np.ndarray:
        images = image[np.newaxis, ...]
        request = pb_utils.InferenceRequest(
            model_name=self._names["preprocessing"],
            requested_output_names=["output"],
            inputs=[pb_utils.Tensor("image", images)],
        )
        resp = request.exec()
        self._raise_for_error(resp)

        output = pb_utils.get_output_tensor_by_name(resp, "output")
        return output.as_numpy().astype(np.float32)

    def _classify(self, images: np.ndarray) -> np.ndarray:
        request = pb_utils.InferenceRequest(
            model_name=self._names["classifier"],
            requested_output_names=["output"],
            inputs=[pb_utils.Tensor("image", images)],
        )
        resp = request.exec()
        self._raise_for_error(resp)

        output = pb_utils.get_output_tensor_by_name(resp, "output")
        logits = output.as_numpy()
        probs = np.exp(logits) / np.exp(logits).sum(1)
        return probs

    def _raise_for_error(self, inference_response):
        if inference_response.has_error():
            raise pb_utils.TritonModelException(inference_response.error().message())



"""Data transforms on top of fiftyone datasets.
"""
import shutil
from pathlib import Path
from typing import Optional

import fiftyone as fo
from fiftyone.types import ImageClassificationDirectoryTree
from PIL import Image, ImageOps, UnidentifiedImageError
from tqdm import tqdm

from ..utils import types
from ..utils.dataset import create_fiftyone_dataset, load_fiftyone_dataset
from ..utils.general import parse_list_str
from ..utils.os_utils import read_json, read_yaml, write_json


def _export_patches(
    dataset: fo.Dataset,
    label_field: str,
    export_dir: Path,
    splits: Optional[list[str]] = None,
) -> None:
    label_type = dataset.get_field(label_field)
    if label_type is None:
        raise KeyError(f"{label_field=} does not exist in {dataset.name=}")
    label_type = label_type.document_type
    if label_type == fo.Classification:
        patches = dataset.exists(label_field)
    elif label_type in [fo.Detections, fo.Polylines]:
        patches = dataset.to_patches(label_field)
    else:
        raise ValueError(f"{label_type=} cannot be exported as patches")
    if splits:
        for tag in splits:
            patches.match_tags(tag).export(
                str(export_dir / tag),
                dataset_type=ImageClassificationDirectoryTree,
                label_field=label_field,
            )
    else:
        patches.export(
            str(export_dir),
            dataset_type=ImageClassificationDirectoryTree,
            label_field=label_field,
        )


def to_patches(
    dataset: str,
    label_field: str | list[str],
    to_name: str,
    export_dir: str,
    overwrite: bool = False,
    splits: Optional[list[str]] = None,
    **kwargs,
) -> fo.Dataset:
    """Crop out patches from a dataset and create a new one.

    Args:
        dataset: a fiftyone dataset with detections
        label_field: label field(s) with detection, classification or polylines
        to_name: a new dataset name for patches
        export_dir: where to save crops
        overwrite: if True and that name already exists, delete it
        splits: if provided, these tags will be used to split patches into subsets
        **kwargs: dataset filters

    Returns:
        fiftyone dataset object
    """
    export_dir = Path(export_dir)

    # prompt overwriting if dataset or folder exist
    if not overwrite:
        if fo.dataset_exists(to_name):
            raise ValueError(
                f"{to_name=} dataset already exists. Use --overwrite or delete it."
            )
        if export_dir.exists():
            raise ValueError(
                f"{str(export_dir)=} already exists. "
                "User --overwrite or delete it manually"
            )
    else:
        if export_dir.exists():
            shutil.rmtree(export_dir)

    dataset = load_fiftyone_dataset(dataset, **kwargs)
    label_field = parse_list_str(label_field)

    # make sure splits are present if given
    if splits:
        splits = parse_list_str(splits)
        tag_counts = dataset.count_sample_tags()
        assert all(
            [s in tag_counts for s in splits]
        ), f"{dataset.name=} does not contain all {splits=}"

    # export each label field
    for field in label_field:
        assert dataset.has_sample_field(
            field
        ), f"{dataset.name=} does not contain {field=}"
        _export_patches(dataset, field, export_dir, splits)

    # import all together, tag if needed
    new = create_fiftyone_dataset(
        name=to_name,
        src=export_dir if splits is None else None,
        dataset_type=ImageClassificationDirectoryTree,
        overwrite=overwrite,
    )
    if splits:
        for tag in splits:
            new.add_dir(
                dataset_dir=str(export_dir / tag),
                dataset_type=ImageClassificationDirectoryTree,
                tags=tag,
            )
    return new


def delete_field(dataset: str, fields: types.LIST_STR_STR):
    """Delete one or more fields from a dataset

    Args:
        dataset: fiftyone dataset name
        fields: fields to delete

    Returns:
        a fiftyone dataset
    """
    dataset = load_fiftyone_dataset(dataset)
    fields = parse_list_str(fields)
    for field in fields:
        dataset.delete_sample_field(field)
        print(f"{field=} deleted from {dataset.name=}")
    return dataset


def prefix_label(dataset: str, label_field: str, dest_field: str, prefix: str):
    """Prepend each label with given prefix

    Args:
        dataset: fiftyone dataset name
        label_field: a field with class labels
        dest_field: a new field to create with '<prefix>_<label>' values
        prefix: a prefix value

    Returns:
        fiftyone dataset object
    """
    dataset = load_fiftyone_dataset(dataset)
    values = [
        fo.Classification(label=f"{prefix}_{smp[label_field].label}")
        for smp in dataset.select_fields(label_field)
    ]
    dataset.set_values(dest_field, values)
    return dataset


def merge_diff(
    dataset: str,
    image_dir: str,
    tags: types.LIST_STR_STR = None,
    recursive: bool = True,
):
    """Merge new files into an existing dataset.

    Existing files will be skipped.
    No labels for new files are expected.
    Merger happens based on an absolute filepath.

    Args:
        dataset: existing fiftyone dataset
        image_dir: a folder with new files
        tags: tag new samples
        recursive: search for files in subfolders as well

    Returns:
        an updated fiftyone dataset
    """
    dataset = load_fiftyone_dataset(dataset)
    second = fo.Dataset.from_images_dir(image_dir, tags=tags, recursive=recursive)
    dataset.merge_samples(second, skip_existing=True)
    return dataset


def delete_samples(dataset: str, **kwargs):
    """Delete samples and associated files from a dataset

    Args:
        dataset: fiftyone dataset name
        **kwargs: dataset filters to select samples for deletion
            (must be provided)

    Returns:
        None
    """
    assert bool(kwargs), "Danger: provide dataset filters to select a subset"
    subset = load_fiftyone_dataset(dataset, **kwargs)
    delete_ids = []
    for smp in subset.select_fields(["id", "filepath"]):
        Path(smp.filepath).unlink()
        delete_ids.append(smp.id)

    full_dataset = fo.load_dataset(dataset)
    full_dataset.delete_samples(delete_ids)
    print(f"{len(delete_ids)} files deleted and removed from {dataset=}")


def exif_transpose(dataset: str, **kwargs):
    """Rotate images that have a PIL rotate tag

    Args:
        dataset: fiftyone dataset name
        **kwargs: dataset loading filters

    Returns:
        None
    """
    dataset = load_fiftyone_dataset(dataset, **kwargs)
    for smp in tqdm(dataset.select_fields("filepath"), desc="transposing"):
        try:
            orig = Image.open(smp.filepath)
            transposed = ImageOps.exif_transpose(orig)
            transposed.save(smp.filepath)
        except UnidentifiedImageError as e:
            print(e, "at", smp.filepath)


def map_labels(
    dataset: str,
    from_field: str,
    to_field: str,
    label_mapping: Optional[dict] = None,
    overwrite: bool = False,
    **kwargs,
) -> fo.DatasetView:
    """Create a new dataset field with mapped labels.

    Args:
        dataset: fiftyone dataset name
        from_field: source label field
        to_field: a new label field
        label_mapping: label mapping (use {}/None for creating a field copy)
        overwrite: if to_field already exists, then overwrite it
        **kwargs: dataset loading kwargs

    Returns:
        dataset view
    """
    dataset = load_fiftyone_dataset(dataset, **kwargs)

    if overwrite and dataset.has_sample_field(to_field):
        delete_field(dataset.dataset_name, to_field)
    elif not overwrite and dataset.has_sample_field(to_field):
        raise ValueError(f"{to_field=} already exists")

    dataset.clone_sample_field(from_field, to_field)
    if bool(label_mapping):
        dataset = dataset.map_labels(to_field, label_mapping)
        dataset.save(to_field)
    return dataset


def _update_labels(labels: fo.Detections, new_label: str):
    for one in labels.detections:
        one.label = new_label


def from_labels(dataset: str, label_field: str, from_field: str, **kwargs):
    """Re-assign classification label to detection labels.

    Args:
        dataset: fiftyone dataset name
        label_field: a field with detections to be updated
        from_field: a field with classification to get labels from
        **kwargs: dataset loading filters
    """
    dataset = load_fiftyone_dataset(dataset, **kwargs)
    dataset = dataset.exists(label_field)

    assert dataset.has_sample_field(
        label_field
    ), f"Dataset does not contain {label_field=}."
    assert (
        doc_type := dataset.get_field(label_field).document_type
    ) == fo.Detections, f"{label_field=} has to be of type Detections, got {doc_type=}."
    assert dataset.has_sample_field(
        from_field
    ), f"Dataset does not contain {from_field=}."
    assert (
        doc_type := dataset.get_field(from_field).document_type
    ) == fo.Classification, (
        f"{from_field=} has to be of type Detections, got {doc_type=}."
    )

    for smp in tqdm(dataset.select_fields([label_field, from_field])):
        _update_labels(smp[label_field], smp[from_field].label)
        smp.save()


def from_label_tag(dataset: str, label_field: str, label_tag: str, **kwargs) -> dict:
    """Update a label_field label with its label_tag.

    Args:
        dataset: fiftyone dataset name
        label_field: a field that contains detections labels.
        label_tag: labels that contain this tag, will be renamed to it.
        **kwargs: dataset loading filters

    Returns:
        updated label values
    """
    kwargs = kwargs | {"label_tags": label_tag}
    dataset = load_fiftyone_dataset(dataset, **kwargs)

    for smp in tqdm(dataset.select_fields(label_field), desc="updating samples"):
        for det in smp[label_field].detections:
            if label_tag in det.tags:
                det.label = label_tag
                smp.save()

    return dataset.count_values(f"{label_field}.detections.label")


def combine_datasets(
    dest_name: str,
    label_field: str,
    cfg: str,
    persistent: bool = True,
    overwrite: bool = False,
):
    """Create a new dataset by adding samples from multiple datasets.

    List of datasets and filters are specified in a yaml config file.
    Source label fields will be renamed to a destination label field.

    Args:
        dest_name: a new dataset name
        label_field: a new label field
        cfg: path to yaml config
        persistent: whether to persist destination dataset (False for testing)
        overwrite: if dataset exists, overwrite it

    Returns:
        a dataset instance
    """
    cfg = read_yaml(cfg)
    assert "datasets" in cfg and isinstance(dataset_cfg := cfg["datasets"], list)
    assert len(dataset_cfg) > 0

    dataset = create_fiftyone_dataset(
        dest_name, src=None, persistent=persistent, overwrite=overwrite
    )
    for one in dataset_cfg:
        assert "name" in one and isinstance(one["name"], str)
        if "filters" in one:
            assert isinstance(one["filters"], dict)
        else:
            one["filters"] = {}
        assert "label_field" in one and isinstance(one["label_field"], str)

        temp_name = f"{dest_name}_{one['name']}"
        temp = load_fiftyone_dataset(one["name"], **one["filters"]).clone(
            name=temp_name, persistent=False
        )
        temp.clone_sample_field(one["label_field"], label_field)
        if "tags" in one:
            temp.tag_samples(one["tags"])

        dataset.add_samples(temp.select_fields([label_field, "tags"]))
        fo.delete_dataset(temp_name)

    return dataset


def fix_filepath(src: str, from_dir: str, to_dir: str) -> None:
    """Replace from_dir part to to_dir in each sample's filepath in samples.json file.

    Samples.json file is updated inplace.

    Args:
        src: sample.json file export for fiftyone.types.FiftyOneDataset
        from_dir: relative directory to replace
        to_dir: new relative directory
    """
    # TODO test this
    src = Path(src)
    assert src.exists() and src.suffix == ".json"
    samples = read_json(src)

    to_dir = Path(to_dir)

    def fix_path(path):
        return str(to_dir / Path(path).relative_to(from_dir))

    for smp in samples["samples"]:
        smp["filepath"] = fix_path(smp["filepath"])

    write_json(samples, src)


def transpose_images(dataset: str, **kwargs) -> fo.DatasetView:
    """Rotate images 90 degrees.

    Args:
        dataset: fiftyone dataset name
        **kwargs: dataset loading filters

    Returns:
        a dataset view instance
    """
    assert len(kwargs) > 0, "Danger: provide dataset filters"

    dataset = load_fiftyone_dataset(dataset, **kwargs)

    for smp in tqdm(dataset.select_fields("filepath"), desc="transposing"):
        Image.open(smp.filepath).transpose(Image.ROTATE_90).save(smp.filepath)

    return dataset



"""Display various data about datasets.
"""
from finegrained.data.transforms import delete_field
from finegrained.utils import types
from finegrained.utils.dataset import get_unique_labels, load_fiftyone_dataset
from finegrained.utils.general import find_diff


def print_labels(dataset: str, label_field: str, **kwargs) -> None:
    """Print all classes in the dataset.

    Args:
        dataset: fiftyone dataset name
        label_field: field that contains labels
        **kwargs: dataset loading filters
    """
    dataset = load_fiftyone_dataset(dataset, **kwargs)
    labels = get_unique_labels(dataset, label_field)
    print("\n".join(labels))


def classification_report(
    dataset: str,
    predictions: str,
    gt_field: str = "ground_truth",
    cmat: bool = False,
    **kwargs,
):
    """Print classification report.

    Args:
        dataset: fiftyone dataset name
        predictions: a field with predictions
        gt_field: a field with ground truth labels
        cmat: if True, plot a confusion matrix
        **kwargs: dataset loading filters

    Returns:
        none
    """
    dataset = load_fiftyone_dataset(dataset, **kwargs)
    labels = dataset.distinct(f"{gt_field}.label")
    results = dataset.evaluate_classifications(
        predictions, gt_field=gt_field, classes=labels
    )
    results.print_report()
    if cmat:
        cm = results.plot_confusion_matrix(backend="matplotlib")
        cm.show()


def label_diff(
    dataset: str,
    label_field: str,
    tags_left: types.LIST_STR_STR,
    tags_right: types.LIST_STR_STR,
):
    """Compute difference between two sets of labels.

    Args:
        dataset: fiftyone dataset name
        label_field: field with labels
        tags_left: list of tags for base list of labels
        tags_right: list of tags for intersection comparison
    """
    # TODO test this
    dataset = load_fiftyone_dataset(dataset)
    assert dataset.has_sample_field(label_field)
    assert len(tags_left) > 0
    assert len(tags_right) > 0

    left_labels = get_unique_labels(dataset.match_tags(tags_left), label_field)
    right_labels = get_unique_labels(dataset.match_tags(tags_right), label_field)

    diff = find_diff(left_labels, right_labels)
    return diff


def compute_area(
    dataset: str,
    field: str = "area",
    average_size: bool = False,
    overwrite_metadata: bool = False,
    overwrite: bool = False,
    **kwargs,
) -> tuple[int, int]:
    """Calculate area of an image based on metadata

    Args:
        dataset: fiftyone dataset name
        field: field where to assign area values
        average_size: if True, calculate (width + height)/2 instead
        overwrite_metadata: whether to overwrite metadata
        overwrite: delete field if already exists
        **kwargs: dataset loading filters

    Returns:
        area bounds
    """
    dataset = load_fiftyone_dataset(dataset, **kwargs)
    if dataset.has_sample_field(field):
        if overwrite:
            delete_field(dataset.name, field)
        else:
            raise ValueError(f"{field=} already exists.")

    dataset.compute_metadata(overwrite=overwrite_metadata)
    for smp in dataset.select_fields("metadata"):
        val = (
            (smp.metadata.width + smp.metadata.height) / 2
            if average_size
            else smp.metadata.width * smp.metadata.height
        )
        smp[field] = val
        smp.save()

    return dataset.bounds(field)



"""Defined type annotations.
"""
from typing import Dict, List, Tuple, Union

LIST_STR = List[str]
LIST_STR_STR = Union[LIST_STR, str]
LIST_DICT = List[Dict]
LIST_TYPE = Union[list, tuple]

DICT_STR_FLOAT = Dict[str, float]
DICT_STR_STR = Dict[str, str]

IMAGE_SIZE = Union[int, Tuple[int, int]]



"""Transformers tokenizer in triton's python backend format
"""
import triton_python_backend_utils as pb_utils
from transformers import AutoTokenizer

# TODO link file instead of downloading
# TODO load from a name
# TODO load names from config


class TritonPythonModel:
    def initialize(self, args):
        self.tokenizer = AutoTokenizer.from_pretrained(
            "symanto/sn-xlm-roberta-base-snli-mnli-anli-xnli"
        )

    def execute(self, requests):
        responses = [self._process_request(one) for one in requests]
        return responses

    def _process_request(self, request):
        texts = pb_utils.get_input_tensor_by_name(request, "text")
        texts = texts.as_numpy().tolist()
        texts = [t[0].decode() for t in texts]

        out = self.tokenizer(texts, padding=True, return_tensors="pt")

        input_ids = pb_utils.Tensor("input_ids", out["input_ids"])
        attention_masks = pb_utils.Tensor("attention_mask", out["attention_mask"])
        resp = pb_utils.InferenceResponse(output_tensors=[input_ids, attention_masks])
        return resp



from . import embed
from .image_classification import ImageClassification, ImageTransform
from .image_metalearn import ImageMetalearn
from .image_selfsupervised import ImageSelfSupervised
from .transformers_base import SentenceEmbeddings
from .yolo import YOLOv5Model, YOLOv5Postprocessing, YOLOv5Preprocessing

__all__ = [
    ImageClassification,
    embed,
    ImageSelfSupervised,
    ImageMetalearn,
    SentenceEmbeddings,
    ImageTransform,
    YOLOv5Model,
    YOLOv5Preprocessing,
    YOLOv5Postprocessing,
]



"""Torch and torchvision utils.
"""
from typing import Optional

import torch
import torchvision.transforms as T
from pytorch_lightning.accelerators import CUDAAccelerator, MPSAccelerator

from finegrained.utils import types

cuda_available = torch.cuda.is_available()


def get_device(device_type: Optional[str] = None) -> tuple[torch.device, int]:
    """Get torch.device and its count with cuda/mps if available, else cpu

    Args:
        device_type: preferred device type

    Returns:
        torch.device instance and its count
    """
    if (
        device_type is None and CUDAAccelerator.is_available()
    ) or device_type == "cuda":
        return torch.device("cuda"), CUDAAccelerator.auto_device_count()
    elif (
        device_type is None and MPSAccelerator.is_available()
    ) or device_type == "mps":
        return torch.device("mps"), MPSAccelerator.auto_device_count()
    else:
        if device_type is not None and device_type != "cpu":
            raise ValueError(f"Unknown device type {device_type}")
        return torch.device("cpu"), 1


def get_default_batch_size() -> int:
    return 16 if cuda_available else 4


def _parse_transform(
    name: str,
    args: Optional[types.LIST_TYPE] = (),
    kwargs: Optional[dict] = {},
) -> list:
    return getattr(T, name)(*args, **kwargs)


def parse_transforms(transforms: types.LIST_DICT, normalize: bool = True) -> T.Compose:
    if transforms is None:
        return None
    transforms = [_parse_transform(**t) for t in transforms]
    if normalize:
        transforms.append(T.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)))
    return T.Compose(transforms)



"""Base logic to use Flash datamodules, models and and trainers.
"""
import itertools
from itertools import chain
from pathlib import Path
from typing import Tuple

import fiftyone as fo
import torch
from flash import Trainer
from flash.core.classification import FiftyOneLabelsOutput
from tqdm import tqdm

from finegrained.models.torch_utils import get_device
from finegrained.utils.dataset import load_fiftyone_dataset
from finegrained.utils.os_utils import read_yaml


def _map_classifications_to_detections(predictions, patches, patch_field) -> dict:
    paired = []
    for clf, smp in zip(predictions, patches):
        detection = fo.Detection(
            label=clf.label,
            bounding_box=smp[patch_field].bounding_box,
            confidence=clf.confidence,
        )
        paired.append((detection, smp.sample_id))

    detections = {}
    for key, group in itertools.groupby(paired, lambda x: x[1]):
        detections[key] = fo.Detections(detections=[d for d, _ in group])
    return detections


class FlashFiftyOneTask:
    """A base class to inherit from for task-specific learners."""

    _model = None
    _data = None
    _prediction_dataset = None
    _patch_dataset = None

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, val):
        self._model = val

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, val):
        self._data = val

    @property
    def train_dataset(self):
        return self._train_dataset

    @train_dataset.setter
    def train_dataset(self, val):
        self._train_dataset = val

    @property
    def prediction_dataset(self):
        return self._prediction_dataset

    @prediction_dataset.setter
    def prediction_dataset(self, val):
        self._prediction_dataset = val

    @property
    def patch_dataset(self):
        return self._patch_dataset

    @patch_dataset.setter
    def patch_dataset(self, val):
        self._patch_dataset = val

    @property
    def data_keys(self):
        return ["dataset", "label_field"]

    @property
    def model_keys(self):
        return ["backbone"]

    @property
    def trainer_keys(self):
        return ["epochs"]

    def _init_training_datamodule(self, *args, **kwargs):
        raise NotImplementedError("Subclass has to implement this method")

    def _init_prediction_datamodule(self, *args, **kwargs):
        raise NotImplementedError("Subclass has to implement this method")

    def _init_model(self, *args, **kwargs):
        raise NotImplementedError("Subclass has to implement this method")

    def _load_pretrained_model(self, *args, **kwargs):
        raise NotImplementedError("Subclass has to implement this method")

    def _load_fiftyone_dataset(self, name: str, **kwargs):
        dataset = load_fiftyone_dataset(name, **kwargs)
        return dataset

    @staticmethod
    def calculate_features(model, dataloader) -> Tuple[torch.Tensor, torch.Tensor]:
        """Extract embeddings for all samples in a dataloader

        Args:
            model: flash model instance
            dataloader: torchvision dataloader instance

        Returns:
            a tuple with features and labels if present
        """
        feature_extractor = model.eval()
        features = []
        labels = []

        with torch.no_grad():
            for batch in tqdm(dataloader, desc="extracting features"):
                features.append(feature_extractor(batch["input"]))
                if "target" in batch.keys():
                    labels.append(batch["target"].argmax(dim=1))

        features = torch.cat(features)
        labels = torch.cat(labels) if len(labels) > 0 else labels
        return features, labels

    def _validate_train_config(self, cfg: dict):
        self.cfg_keys = dict(
            data=self.data_keys,
            model=self.model_keys,
            trainer=self.trainer_keys,
        )

        for key in self.cfg_keys.keys():
            assert key in cfg, f"config file has to contain {key=}"
        for comp, comp_keys in self.cfg_keys.items():
            for key in comp_keys:
                assert key in cfg[comp], f"{comp} has to contain {key=}"

    def _get_available_backbones(self):
        raise NotImplementedError("Subclass has to implement this method")

    def _finetune(self, epochs: int, **kwargs):
        assert self.model is not None, "make sure a model is initialized"
        assert self.data is not None, "make sure a datamodule is initialized"

        device, device_count = get_device(kwargs.get("device"))
        trainer = Trainer(
            max_epochs=epochs,
            limit_train_batches=kwargs.get("limit_train_batches"),
            limit_val_batches=kwargs.get("limit_val_batches"),
            accelerator=device.type,
            devices=device_count,
        )

        strategy = kwargs.get("strategy", ("freeze_unfreeze", 1))
        if isinstance(strategy, list):
            strategy = tuple(strategy)

        trainer.finetune(
            self.model,
            datamodule=self.data,
            strategy=strategy,
        )

        trainer.save_checkpoint(kwargs.get("save_checkpoint", "model.pt"))

        if hasattr(self, "test_dataset") and (
            test_loader := self.data.test_dataloader()
        ):
            trainer.test(dataloaders=test_loader)

    def _predict(self):
        device, device_count = get_device()
        trainer = Trainer(accelerator=device.type, devices=device_count)
        predictions = trainer.predict(
            self.model,
            datamodule=self.data,
            output=FiftyOneLabelsOutput(self.model.labels, return_filepath=False),
        )
        predictions = list(chain.from_iterable(predictions))
        return predictions

    def finetune(self, cfg: str | Path):
        """Fine-tune model with given data

        Args:
            cfg: path to yaml file with data, model and trainer configs
        """
        cfg = read_yaml(cfg)
        self._validate_train_config(cfg)
        self._init_training_datamodule(**cfg["data"])
        self._init_model(**cfg["model"])
        self._finetune(**cfg["trainer"])

    def predict(
        self,
        dataset: str,
        label_field: str,
        ckpt_path: str,
        image_size: Tuple[int, int] = (224, 224),
        batch_size: int = 4,
        patch_field: str = None,
        **kwargs,
    ):
        """Classify samples from a dataset and assign values to label field

        Args:
            dataset: which dataset to run samples on
            label_field: which field to assign predictions to
            ckpt_path: flash model checkpoint path
            image_size: Image size for inference
            batch_size: predictions batch size
            patch_field: run predictions on patches on this field
            **kwargs: dataset loading filters

        Returns:
            none
        """
        self._init_prediction_datamodule(
            dataset,
            image_size=image_size,
            batch_size=batch_size,
            patch_field=patch_field,
            **kwargs,
        )
        assert (
            self.prediction_dataset is not None
        ), "Make sure _init_prediction_datamodule() creates self.prediction_dataset"
        self._load_pretrained_model(ckpt_path)
        predictions = self._predict()
        if bool(patch_field):
            self.patch_dataset.set_values(label_field, predictions)

            for smp in self.prediction_dataset:
                detections = []
                for det in smp[patch_field].detections:
                    clf = self.patch_dataset[det.patch_filepath][label_field]
                    detections.append(
                        fo.Detection(
                            label=clf.label,
                            bounding_box=det.bounding_box,
                            confidence=clf.confidence,
                        )
                    )

                smp[label_field] = fo.Detections(detections=detections)
                smp.save()
        else:
            self.prediction_dataset.set_values(label_field, predictions)

    def list_backbones(self, prefix: str = None):
        backbones = self._get_available_backbones()
        if prefix:
            backbones = list(filter(lambda x: x.startswith(prefix), backbones))
        return backbones



"""Log metrics and models to MLflow.
"""
import os
import warnings
from pathlib import Path
from typing import Optional

import mlflow
import onnx
from urllib3.exceptions import InsecureRequestWarning

from finegrained.utils import mlflow_server
from finegrained.utils.mlflow_server import get_tensorboard_files
from finegrained.utils.os_utils import read_file_config, read_yaml

warnings.filterwarnings("once", category=InsecureRequestWarning)


def log_events(path: str | list) -> None:
    """Log events history as metrics to MLflow

    Args:
        path: tensorboard events path
    """
    if isinstance(path, str):
        path = [path]

    for one in path:
        scalars = mlflow_server.read_tensorboard_scalars(one)
        metrics = mlflow_server.parse_tensorboard_scalars(scalars)

        for step, vals in enumerate(metrics):
            step = vals.pop("epoch", step)
            mlflow.log_metrics(vals, step=step)


def log_hparams(path: str) -> None:
    """Log hyperparameters to MLflow

    Args:
        path: path to hparams.yaml file
    """
    hparams = read_yaml(path)
    labels = hparams.pop("labels")

    mlflow.log_params(hparams)
    if bool(labels):
        mlflow.log_text("\n".join(labels), artifact_file="labels.txt")


def log_ckpt(path: str) -> None:
    """Log a torch checkpoint to MLflow

    Args:
        path: path to a checkpoint file
    """
    path = Path(path)
    if path.is_file():
        mlflow.log_artifact(str(path), "checkpoints")
    else:
        mlflow.log_artifacts(str(path), "checkpoints")


def log_run(
    name: Optional[str] = None,
    events: Optional[str] = None,
    hparams: Optional[str] = None,
    ckpt: Optional[str] = None,
    model: Optional[str] = None,
    metrics: dict = {},
    log_dir: Optional[str] = None,
    tracking_uri: str = "./mlruns",
    experiment_name: Optional[str] = None,
    artifact_location: Optional[str] = None,
    env: Optional[str] = None,
):
    """Connect to MLflow and log metrics, params and files.

    Either pass paths to events, hparams and ckpt or a single path to log_dir.

    Args:
        name: mlflow run name
        events: path to tensorboard events file as metrics
        hparams: path to hparams.yaml file as params and labels
        ckpt: path to torch checkpoint file or a dir with checkpoint files
        model: path to onnx model to log as MLflow models
        metrics: pass extra metrics to log
        log_dir: provide path to tensorboard log dir in order to automatically
            get events, hparams and ckpt paths.
        tracking_uri: mlflow tracking uri
        experiment_name: if set, use this experiment name
            (if does not exist, creates a new one)
        artifact_location: set artifact location for this experiment
                            (applicable only when a new experiment created)
        env: path to a file with environment variables for MLflow S3

    Returns:
        MLflow run id
    """

    mlflow_server.connect_to_mlflow(
        tracking_uri,
        experiment_name=experiment_name,
        artifact_location=artifact_location,
    )
    if env:
        env = read_file_config(env)
        for k, v in env.items():
            os.environ[k] = v

    if log_dir:
        events, hparams, ckpt = get_tensorboard_files(log_dir)

    with mlflow.start_run(run_name=name) as run:
        run_id = run.info.run_id

        if events:
            log_events(events)
        if hparams:
            log_hparams(hparams)
        if ckpt:
            log_ckpt(ckpt)
        if model:
            log_model(model)
        if metrics:
            mlflow.log_metrics(metrics)

    return run_id


def log_model(path: str) -> None:
    """Log ONNX model to MLflow run.

    Args:
        path: path to onnx model file
    """
    assert path.endswith(".onnx")
    model = onnx.load_model(path)
    mlflow.onnx.log_model(model, artifact_path="model")



"""Utils to work with structured data.
"""
from pathlib import Path
from typing import Optional, Union

import numpy as np
import pandas as pd


def load_data(
    path: str, key: Optional[str] = None, **kwargs
) -> Union[pd.DataFrame, np.ndarray, list]:
    """Load data from csv/json file."""
    if isinstance(path, list):
        data = pd.DataFrame(path)
    else:
        path = Path(path)
        if path.suffix == ".csv":
            data = pd.read_csv(path, **kwargs)
        elif path.suffix == ".csv":
            data = pd.read_json(path, **kwargs)
        elif path.suffix == ".npy":
            data = np.load(path, allow_pickle=False)
        else:
            raise ValueError(f"Unsupported file type: {path.suffix}")

    if key:
        data = data[key].tolist()

    return data


def write_data(path: str, data: list):
    """Write data to csv/json file."""
    path = Path(path)
    data = pd.DataFrame(data)
    if path.suffix == ".csv":
        data.to_csv(path, index=False)
    elif path.suffix == ".json":
        data.to_json(path, index=False)
    elif path.suffix == ".npy":
        with open(path, "wb") as f:
            np.save(f, data, allow_pickle=False)
    else:
        raise ValueError(f"Unsupported file type: {path.suffix}")



"""Train, eval and predict with image embedding model.
"""
from flash.image import ImageEmbedder

from finegrained.models import ImageClassification


class ImageSelfSupervised(ImageClassification):
    """Use unsupervised learning in order to classify images"""

    @property
    def data_keys(self):
        return ["dataset"]

    def _init_model(
        self,
        backbone: str,
        training_strategy: str = "simclr",
        head: str = "simclr_head",
        pretraining_transform: str = "simclr_transform",
        **kwargs
    ):
        self.model = ImageEmbedder(
            training_strategy=training_strategy,
            head=head,
            pretraining_transform=pretraining_transform,
            backbone=backbone,
            pretrained=kwargs.get("pretrained", True),
            optimizer=kwargs.get("optimizer", "Adam"),
            learning_rate=kwargs.get("learning_rate", 0.001),
        )

    def _get_available_backbones(self):
        return ImageEmbedder.available_backbones()



import fiftyone as fo
from fiftyone import ViewField as F
from fiftyone.types import ImageClassificationDirectoryTree

from finegrained.utils import types
from finegrained.utils.general import parse_list_str
from finegrained.utils.types import LIST_STR


def load_fiftyone_dataset(
    dataset: str,
    include_labels: dict[str, LIST_STR] = {},
    exclude_labels: dict[str, LIST_STR] = {},
    include_tags: LIST_STR = None,
    exclude_tags: LIST_STR = None,
    max_samples: int = None,
    label_conf: float = 0.0,
    fields_exist: LIST_STR = None,
    not_exist: LIST_STR = None,
    label_tags: LIST_STR = None,
) -> fo.DatasetView | fo.Dataset:
    """Load a dataset and apply view filters

    Args:
        dataset: fiftyone dataset name
        include_labels: keep samples that have fields with these values
        exclude_labels: exclude these labels (not samples) from dataset view
        include_tags: keep samples that match these sample tags,
                        or use 'all' to solely apply a confidence threshold
        exclude_tags: exclude samples that match these sample tags
        max_samples: randomly select this number of samples if specified
        label_conf: if 'include_labels' specified, apply confidence threshold
            (only works when include_labels is specified)
        fields_exist: keep samples that contain these fields
        not_exist: keep samples that do Not contain these fields
        label_tags: keep samples and labels that match these label tags

    Returns:
        a fiftyone dataset
    """
    dataset = fo.load_dataset(dataset)

    # tags
    if bool(include_tags):
        dataset = dataset.match_tags(include_tags)
    if bool(exclude_tags):
        dataset = dataset.match_tags(exclude_tags, False)

    # labels
    if bool(include_labels):
        for field, values in include_labels.items():
            if label_conf > 0:
                dataset = dataset.filter_labels(field, F("confidence") >= label_conf)
            if values != "all":  # TODO explain in docs
                filter_fn = F("label").is_in(parse_list_str(values))
                dataset = dataset.filter_labels(field, filter_fn)
    if bool(exclude_labels):
        for field, values in exclude_labels.items():
            filter_fn = ~F("label").is_in(parse_list_str(values))
            dataset = dataset.filter_labels(field, filter_fn, False)

    # fields
    if bool(fields_exist):
        for field in parse_list_str(fields_exist):
            dataset = dataset.exists(field)
    if bool(not_exist):
        for field in parse_list_str(not_exist):
            dataset = dataset.exists(field, False)

    # label tags
    if bool(label_tags):
        dataset = dataset.match_labels(tags=parse_list_str(label_tags))

    # max samples
    if bool(max_samples):
        dataset = dataset.take(min(max_samples, len(dataset)))

    assert len(dataset) > 0, "No samples left after filtering the dataset"
    return dataset


def create_fiftyone_dataset(
    name: str,
    src: str,
    dataset_type=ImageClassificationDirectoryTree,
    overwrite: bool = False,
    persistent: bool = True,
) -> fo.Dataset:
    """Create a fiftyone dataset from a directory (or an empty dataset)

    Args:
        name: a name for this dataset
        src: dataset directory (if None, an empty dataset will be created)
        dataset_type: fiftyone dataset format
        overwrite: whether to overwrite if that name is already taken
        persistent: whether to persist this dataset in mongo

    Returns:
        a fiftyone dataset
    """
    if (exists := fo.dataset_exists(name)) and not overwrite:
        raise ValueError(
            f"Dataset {name!r} already exists!. User overwrite=True to delete it"
        )
    elif overwrite and exists:
        fo.delete_dataset(name, verbose=True)

    if src:
        dataset = fo.Dataset.from_dir(
            dataset_dir=src,
            dataset_type=dataset_type,
            name=name,
        )
    else:
        dataset = fo.Dataset(name)

    if persistent:
        dataset.persistent = True
    return dataset


def get_unique_labels(dataset: fo.Dataset, label_field: str) -> list[str]:
    """Extract a list of labels from a field

    Args:
        dataset: a fiftyone dataset object
        label_field: a field to extract labels from

    Returns:
        a list of class labels
    """
    label_fields = parse_list_str(label_field)
    labels = []
    for field in label_fields:
        if not dataset.has_sample_field(field):
            raise KeyError(f"{dataset.name} does not contain {field=}")
        label_type = dataset.get_field(field).document_type
        if label_type == fo.Classification:
            view_field = f"{field}.label"
        elif label_type == fo.Detections:
            view_field = f"{field}.detections.label"
        elif label_type == fo.Polylines:
            view_field = f"{field}.polylines.label"
        else:
            raise NotImplementedError(f"Not implemented for {label_type=}.")
        labels.extend(dataset.distinct(F(view_field)))
    return labels


def get_all_filepaths(dataset: fo.Dataset) -> types.LIST_STR:
    """Retrieve all filepaths from a sample collection

    Args:
        dataset: fiftyone dataset object or sample collection

    Returns:
        a list of absolute paths
    """
    files = [smp.filepath for smp in dataset.select_fields("filepath")]
    return files



"""YOLO utils for training, evaluating and exporting.
"""
import shutil
from typing import List, Tuple

import torch
import torchvision.transforms as T
import torchvision.transforms.functional as F
from torch import nn
from torchvision.ops import batched_nms

from finegrained.utils.triton import TritonExporter

DEFAULTS = dict(iou_threshold=0.4, conf_threshold=0.25, stride=32, max_detections=100)


def _get_dim(t: torch.Tensor) -> torch.Tensor:
    """get a size without casting to python"""
    bools = t.greater_equal(0)
    return bools.sum(0, keepdim=True)


def get_tensor_size(x: torch.Tensor) -> torch.Tensor:
    """return (h,w) of a tensor as a tensor"""
    # return torch.tensor([x.size(1), x.size(2)])
    h = _get_dim(x[0, :, 0])
    w = _get_dim(x[0, 0, :])
    return torch.cat([h, w])


def apply_conf_threshold(
    boxes: torch.Tensor,
    scores: torch.Tensor,
    labels: torch.Tensor,
    label_probs: torch.Tensor,
    threshold: float,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """Filter out low-confidence results.


    :param scores: scores to apply the threshold to
    :param threshold: a threshold value
    :return: a list of filtered tensors
    """
    above_conf = scores > threshold
    boxes = boxes[above_conf]
    scores = scores[above_conf]
    labels = labels[above_conf]
    label_probs = label_probs[above_conf]
    return boxes, scores, labels, label_probs


def scale_coords(
    img1_shape: torch.Tensor, coords: torch.Tensor, img0_shape: torch.Tensor
) -> torch.Tensor:
    """Rescale coords (xyxy) from img1_shape to img0_shape"""
    gain = (img1_shape / img0_shape).min()
    pad = (img1_shape - img0_shape * gain) / 2
    coords[:, [0, 2]] -= pad[1]  # x padding
    coords[:, [1, 3]] -= pad[0]  # y padding
    coords[:, :4] /= gain
    clip_coords(coords, img0_shape)
    return coords


def clip_coords(boxes: torch.Tensor, shape: torch.Tensor) -> None:
    """Clip bounding xyxy bounding boxes to image shape (height, width)"""
    boxes[:, 0].clamp_(0, shape[1])  # x1
    boxes[:, 1].clamp_(0, shape[0])  # y1
    boxes[:, 2].clamp_(0, shape[1])  # x2
    boxes[:, 3].clamp_(0, shape[0])  # y2


def xyxy2xywhn(x: torch.Tensor, w: int = 640, h: int = 640) -> torch.Tensor:
    """Convert nx4 boxes from [x1, y1, x2, y2] to [x, y, w, h]
    normalized where xy1=top-left, xy2=bottom-right"""
    y = x.clone()
    y[:, 0] = ((x[:, 0] + x[:, 2]) / 2) / w  # x center
    y[:, 1] = ((x[:, 1] + x[:, 3]) / 2) / h  # y center
    y[:, 2] = (x[:, 2] - x[:, 0]) / w  # width
    y[:, 3] = (x[:, 3] - x[:, 1]) / h  # height
    return y


def rescale_boxes(
    boxes: torch.Tensor, input_size: torch.Tensor, img_size: torch.Tensor
) -> torch.Tensor:
    """Scale detected boxes to orig image size
    and normalize to 0-1 range in the xywh format.
    """
    boxes = scale_coords(img1_shape=input_size, coords=boxes, img0_shape=img_size)
    h, w = img_size[0], img_size[1]
    boxes = xyxy2xywhn(boxes, w=w, h=h)
    return boxes


def xywh2xyxy(x: torch.Tensor) -> torch.Tensor:
    """Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2]
    where xy1=top-left, xy2=bottom-right"""
    y = x.clone()
    y[:, 0] = x[:, 0] - x[:, 2] / 2  # top left x
    y[:, 1] = x[:, 1] - x[:, 3] / 2  # top left y
    y[:, 2] = x[:, 0] + x[:, 2] / 2  # bottom right x
    y[:, 3] = x[:, 1] + x[:, 3] / 2  # bottom right y
    return y


def parse_yolov5_prediction(
    prediction: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """Parse raw predictions into boxes, scores, label probs and indices."""
    xywh = prediction[:, :4]
    obj_scores = prediction[:, 4]
    label_probs = prediction[:, 5:]
    clf_scores, labels = label_probs.max(dim=1)
    boxes = xywh2xyxy(xywh)
    scores = obj_scores * clf_scores
    return boxes, scores, labels, label_probs


class Letterbox(nn.Module):
    """Letterbox transformation for object detection.

    Resize larger size to new_shape if size is bigger than new_shape
    or pad to the closest divisible of stride if less than new_shape.
    Resize smaller size with the same ratio and
    pad it to the closest divisible by stride.

    :param new_shape: new max size, will result in (<=new_shape, <=new_shape)
    :param stride: both dimensions will be divisible by this number
    """

    def __init__(self, new_shape: int, stride: int):
        super(Letterbox, self).__init__()
        self.new_shape = nn.Parameter(torch.tensor(new_shape), requires_grad=False)
        self.stride = nn.Parameter(torch.tensor(stride), requires_grad=False)

    def forward(self, sample: torch.Tensor) -> torch.Tensor:
        shape: torch.Tensor = get_tensor_size(sample)
        scale_ratio: torch.Tensor = (self.new_shape / shape.max()).clip(0.0, 1.0)
        new_unpad: torch.Tensor = (shape * scale_ratio).round().to(torch.long)  # h,w
        delta_hw: torch.Tensor = (
            torch.fmod(self.new_shape - new_unpad, self.stride) / 2
        ).to(torch.long)
        new_unpad_l: List[torch.Tensor] = [new_unpad[0], new_unpad[1]]
        sample = F.resize(
            sample, new_unpad_l, interpolation=F.InterpolationMode.BILINEAR
        )
        padding: List[torch.Tensor] = [
            delta_hw[1],
            delta_hw[0],
            delta_hw[1],
            delta_hw[0],
        ]
        sample = F.pad(sample, padding, fill=sample.to(torch.float).mean().item())
        return sample


class YOLOv5Preprocessing(nn.Module, TritonExporter):
    """Pre-process yolov5 inputs.

    Run a raw input tensor image thru letterbox transformation
    and return a resized and normalized image along with resize ratio
    """

    def __init__(self, image_size, stride=DEFAULTS["stride"]):
        super(YOLOv5Preprocessing, self).__init__()
        self.transforms = nn.Sequential(Letterbox(new_shape=image_size, stride=stride))
        self.reshape = T.Lambda(lambda x: x.permute(2, 0, 1))

    def forward_one(
        self, sample: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        sample = self.reshape(sample)
        img_size = get_tensor_size(sample)
        sample = self.transforms(sample)
        sample = (sample / 255.0).to(torch.float32)
        return sample, img_size, get_tensor_size(sample)

    def forward(
        self, sample: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        out, img_size, input_size = self.forward_one(sample)
        return out.unsqueeze(0), img_size, input_size

    def _load_model_torch(self, *args, **kwargs) -> torch.nn.Module:
        return self

    def generate_dummy_inputs(self, *args, **kwargs) -> List[torch.Tensor]:
        return (torch.randint(0, 255, size=(600, 680, 3), dtype=torch.uint8),)

    @property
    def output_names(self):
        return ["output", "image_size", "input_size"]

    @property
    def triton_batch_size(self):
        return 0

    @property
    def dynamic_axes(self):
        return {
            "image": {0: "h", 1: "w"},
            "output": {2: "h", 3: "w"},
            # "image_size": {0: "batch"},
            # "input_size": {0: "batch"},
        }


class YOLOv5Postprocessing(nn.Module, TritonExporter):
    """Post-process detection outputs."""

    def __init__(
        self,
    ):
        super(YOLOv5Postprocessing, self).__init__()
        self._ = nn.Sequential()
        self.conf_threshold: float = DEFAULTS["conf_threshold"]
        self.iou_threshold: float = DEFAULTS["iou_threshold"]
        self.max_detections: float = DEFAULTS["max_detections"]

    def _filter_detections(self, detections: torch.Tensor):
        """Parse detections, apply conf. threshold and run NMS."""
        boxes, scores, labels, label_probs = parse_yolov5_prediction(detections)
        boxes, scores, labels, label_probs = apply_conf_threshold(
            boxes=boxes,
            scores=scores,
            labels=labels,
            label_probs=label_probs,
            threshold=self.conf_threshold,
        )
        keep_inds: torch.Tensor = batched_nms(
            boxes, scores, labels, self.iou_threshold
        )[: self.max_detections]
        return boxes[keep_inds], scores[keep_inds].unsqueeze(1), label_probs[keep_inds]

    def _prepare_output(
        self,
        boxes: torch.Tensor,
        scores: torch.Tensor,
        label_probs: torch.Tensor,
        input_size: torch.Tensor,
        img_size: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """rescale bboxes to original image size and one-hot encode labels"""
        rescaled_boxes = rescale_boxes(boxes, input_size=input_size, img_size=img_size)
        return rescaled_boxes, scores, label_probs

    def forward(
        self,
        prediction: torch.Tensor,
        image_size: torch.Tensor,
        input_size: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:

        detections = prediction.squeeze(0)
        # image_size = image_size[0]
        # input_size = input_size[0]

        boxes, scores, label_probs = self._filter_detections(detections)
        out = self._prepare_output(boxes, scores, label_probs, input_size, image_size)
        return out

    def _load_model_torch(self, *args, **kwargs) -> torch.nn.Module:
        return self

    @property
    def input_names(self):
        return ["prediction", "image_size", "input_size"]

    @property
    def output_names(self):
        return ["boxes__0", "scores__1", "class_probs__2"]

    @property
    def dynamic_axes(self):
        return {
            self.input_names[0]: {1: "n_detections", 2: "n_classes"},
            # self.input_names[1]: {0: "batch"},
            # self.input_names[2]: {0: "batch"},
            self.output_names[0]: {0: "n_detections"},
            self.output_names[1]: {0: "n_detections"},
            self.output_names[2]: {0: "n_detections", 2: "n_classes"},
        }

    @property
    def triton_batch_size(self):
        return 0

    def generate_dummy_inputs(self, **kwargs) -> List[torch.Tensor]:
        n = 18900
        prediction = torch.rand(1, n, 10)
        img_size = torch.tensor([480, 640])
        inp_size = torch.tensor([240, 320])
        return prediction, img_size, inp_size

    def _create_triton_config(self, torchscript: bool = False, **kwargs) -> dict:
        return dict(
            backend="pytorch" if torchscript else "onnxruntime",
            max_batch_size=self.triton_batch_size,
            input=[
                dict(
                    name=self.input_names[0],
                    data_type="TYPE_FP32",
                    dims=[1, -1, -1],
                ),
                dict(name=self.input_names[1], data_type="TYPE_INT64", dims=[2]),
                dict(name=self.input_names[2], data_type="TYPE_INT64", dims=[2]),
            ],
            output=[
                dict(
                    name=self.output_names[0],
                    data_type="TYPE_FP32",
                    dims=[-1, -1],
                ),
                dict(name=self.output_names[1], data_type="TYPE_FP32", dims=[-1, 1]),
                dict(
                    name=self.output_names[2],
                    data_type="TYPE_FP32",
                    dims=[-1, -1],
                ),
            ],
        )


class YOLOv5Model(TritonExporter):
    """Create a triton model and ensemble for a trained yolov5 model."""

    @property
    def triton_batch_size(self):
        return 0

    def export_onnx(self, model_path: str, write_path: str, **kwargs):
        """Copy existing ONNX file to write_path."""
        shutil.copy(model_path, write_path)

    def _create_triton_ensemble_config(
        self,
        preprocessing_name: str,
        model_name: str,
        postprocessing_name: str,
        **kwargs
    ):
        return dict(
            platform="ensemble",
            max_batch_size=self.triton_batch_size,
            input=[dict(name="IMAGE", data_type="TYPE_UINT8", dims=[-1, -1, 3])],
            output=[
                dict(name="BOXES", data_type="TYPE_FP32", dims=[-1, -1]),
                dict(name="SCORES", data_type="TYPE_FP32", dims=[-1, 1]),
                dict(name="CLASS_PROBS", data_type="TYPE_FP32", dims=[-1, -1]),
            ],
            ensemble_scheduling=dict(
                step=[
                    dict(
                        model_name=preprocessing_name,
                        model_version=-1,
                        input_map=dict(image="IMAGE"),
                        output_map=dict(
                            output="PREPROCESSED",
                            image_size="IMAGE_SIZE",
                            input_size="INPUT_SIZE",
                        ),
                    ),
                    dict(
                        model_name=model_name,
                        model_version=-1,
                        input_map=dict(images="PREPROCESSED"),
                        output_map=dict(output="DETECTIONS"),
                    ),
                    dict(
                        model_name=postprocessing_name,
                        model_version=-1,
                        input_map=dict(
                            prediction="DETECTIONS",
                            image_size="IMAGE_SIZE",
                            input_size="INPUT_SIZE",
                        ),
                        output_map=dict(
                            boxes__0="BOXES",
                            scores__1="SCORES",
                            class_probs__2="CLASS_PROBS",
                        ),
                    ),
                ]
            ),
        )



"""Generate embeddings for input data.
"""
from typing import Dict, List

from finegrained.models.transformers_base import SentenceEmbeddings
from finegrained.utils.data import load_data, write_data
from finegrained.utils.similarity import SimilaritySearch


def text(model_name: str, input_data: str, input_data_key: str, output: str):
    """Generate embeddings for input data."""
    data = load_data(input_data, key=input_data_key)
    model = SentenceEmbeddings(model_name)
    embeddings = model.embed(data).numpy()
    if output is not None:
        write_data(output, embeddings)
    else:
        return data, embeddings


def most_similar(
    query: str, model_name: str, data: str, embeddings: str, top_k: int = 5
) -> List[Dict]:
    """Find most similar items in the data for the query.

    Args:
        query: Query to find most similar items.
        model_name: which model to embed the query with.
        data: Data to return for most similar items.
        embeddings: Embeddings to use for similarity search.
        top_k: number of top similar to return.

    Returns:
        Most similar items in the data for the query.
    """
    data = load_data(data)
    embeddings = load_data(embeddings)
    sim = SimilaritySearch(data.to_dict("records"), embeddings)
    model = SentenceEmbeddings(model_name)
    top_sim = sim.embed_and_find_similar(query, model, top_k=top_k)
    return top_sim



from finegrained.utils import types


def parse_list_str(value: types.LIST_STR_STR) -> types.LIST_STR:
    """Parses str ot list[str] into a list[str]

    Args:
        value: a str or a list[str]

    Returns:
        a list[str]
    """
    if isinstance(value, str):
        return [value]
    elif isinstance(value, (list, tuple)):
        return value
    else:
        raise TypeError(f"{value=} is not str or list")


def find_diff(left: list, right: list) -> list:
    return list(set(left).difference(right))



"""Dataset converting and exporting utils.
"""
from typing import List, Optional

import fiftyone.types as fot

from finegrained.utils.dataset import get_unique_labels, load_fiftyone_dataset


def to_yolov5(
    dataset: str,
    label_field: str,
    export_dir: str,
    splits: List[str],
    **kwargs,
):
    """Export a dataset into yolov5 format for training

    Args:
        dataset: fiftyone dataset name
        label_field: field that contains labels
        export_dir: where to write data
        splits: which splits to export
        **kwargs: dataset loading filters
    """
    dataset = load_fiftyone_dataset(dataset, **kwargs)
    labels = get_unique_labels(dataset, label_field)
    for tag in splits:
        subset = dataset.match_tags(tag)
        assert len(subset) > 0, f"No samples in the subset with {tag=}"
        subset.export(
            export_dir=export_dir,
            split=tag,
            dataset_type=fot.YOLOv5Dataset,
            label_field=label_field,
            classes=labels,
        )


def to_cvat(dataset: str, label_field: str, export_dir: str, **kwargs):
    """Export a dataset into CVAT format for annotation.

    Args:
        dataset: fiftyone dataset name
        label_field: field that contains labels
        export_dir: where to write data
        **kwargs: dataset loading filters
    """
    dataset = load_fiftyone_dataset(dataset, **kwargs)
    dataset.export(
        export_dir=export_dir,
        dataset_type=fot.CVATImageDataset,
        label_field=label_field,
    )


def to_csv(
    dataset: str,
    label_field: str,
    export_path: str,
    extra_fields: Optional[list[str]] = None,
    **kwargs,
):
    """Export a dataset into CSV format for uploading to external sources.

    Args:
        dataset: fiftyone dataset name
        label_field: field that contains labels (will be mapped to 'label')
        export_path: where to write csv file
        extra_fields: extra fields to be added to csv
        **kwargs: dataset loading filters
    """
    dataset = load_fiftyone_dataset(dataset, **kwargs)
    label_field = f"{label_field}.label"
    fields = {"filepath": "image", label_field: "label"}
    if extra_fields:
        fields.update({k: k for k in extra_fields})
    dataset.export(
        dataset_type=fot.CSVDataset,
        abs_paths=True,
        export_media=False,
        labels_path=export_path,
        fields=fields,
    )



"""Utils to work with transformers package.
"""
import shutil
from pathlib import Path

import torch
from transformers import AutoModel, AutoTokenizer

from finegrained.models import torch_utils
from finegrained.utils import types
from finegrained.utils.triton import init_model_repo, save_triton_config


class MeanPooling(torch.nn.Module):
    def __init__(self):
        super(MeanPooling, self).__init__()
        self._ = torch.nn.Identity()

    def forward(self, model_output, attention_mask):
        token_embeddings = model_output[0]
        sizes = [
            token_embeddings[:, 0, 0].greater_equal(-1e6).sum(),
            token_embeddings[0, :, 0].greater_equal(-1e6).sum(),
            token_embeddings[0, 0, :].greater_equal(-1e6).sum(),
        ]
        input_mask_expanded = attention_mask.unsqueeze(-1).expand(sizes)
        return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(
            input_mask_expanded.sum(1), min=1e-9
        )


class TransformersBase:
    def __init__(
        self,
        model_name: str,
        batch_size: int = torch_utils.get_default_batch_size(),
        device=torch_utils.get_device(),
    ):
        self.device = device
        self._tokenizer = AutoTokenizer.from_pretrained(model_name)
        self._model = AutoModel.from_pretrained(model_name).to(device)
        self.embedding_dim = self._model.embeddings.token_type_embeddings.embedding_dim
        self.model_name = model_name
        self._batch_size = batch_size

    @property
    def batch_size(self):
        return self._batch_size

    def tokenizer(self, inputs: types.LIST_STR):
        tokens = self._tokenizer(
            inputs, padding=True, truncation=True, return_tensors="pt"
        )
        _ = [v.to(self.device) for v in tokens.values()]
        return tokens

    def model(self):
        raise NotImplementedError()


class SentenceEmbeddingsModel(torch.nn.Module):
    def __init__(self, model):
        super().__init__()
        self.model = model
        self.mean_pooling = MeanPooling()

    @torch.no_grad()
    def forward(self, input_ids, attention_mask):
        embeddings = self.model(input_ids, attention_mask)
        embeddings = self.mean_pooling(embeddings, attention_mask)
        return embeddings


class SentenceEmbeddings(TransformersBase):
    """Extract features for text chunks."""

    @torch.no_grad()
    def model(self, tokens: dict) -> torch.Tensor:
        model_output = self._model(**tokens)
        embeddings = MeanPooling().forward(model_output, tokens["attention_mask"])
        return embeddings

    def predict(self, questions):
        if isinstance(questions, str):
            questions = [questions]

        # TODO add batch processing
        tokens = self.tokenizer(questions)
        embeddings = self.model(tokens)
        return embeddings

    def generate_dummy_input(self):
        dummy = ["This is sentence one", "And this is a second sentence"]
        tokens = self.tokenizer(dummy)
        return tokens.data["input_ids"], tokens.data["attention_mask"]

    @property
    def input_names(self):
        return ["input_ids", "attention_mask"]

    @property
    def output_names(self):
        return ["embeddings"]

    @property
    def dynamic_axes(self):
        return {
            "input_ids": {0: "batch_size", 1: "tokens"},
            "attention_mask": {0: "batch_size", 1: "tokens"},
            "embeddings": {0: "batch_size"},
        }

    def export_onnx(self, write_path: str) -> None:
        """Export model into onnx format

        Args:
            write_path:

        Returns:

        """
        # TODO finish with this
        # model = SentenceEmbeddingsModel(self._model.to("cpu"))
        dummy_tokens = self.generate_dummy_input()
        torch.onnx.export(
            self._model,
            dummy_tokens,
            write_path,
            export_params=True,
            input_names=self.input_names,
            output_names=self.output_names,
            dynamic_axes=self.dynamic_axes,
            opset_version=13,
        )

    def export_triton(self, triton_repo: str, triton_name: str, version: int = 1):
        model_version_dir = init_model_repo(triton_repo, triton_name, version)
        self.export_onnx(model_version_dir / "model.onnx")

        config = self._create_triton_config()
        write_config = model_version_dir.parent / "config.pbtxt"
        save_triton_config(config, write_config)

    def _create_triton_config(self):
        return {
            "backend": "onnxruntime",
            "max_batch_size": self.batch_size,
        }

    def _create_ensemble_config(
        self,
    ) -> dict:
        return {
            "platform": "ensemble",
            "max_batch_size": self.batch_size,
            "input": [dict(name="TEXT", data_type="TYPE_STRING", dims=[-1])],
            "output": [
                dict(name="ATTENTION_MASK", data_type="TYPE_INT64", dims=[-1]),
                dict(
                    name="EMBEDDINGS",
                    data_type="TYPE_FP32",
                    dims=[-1, self.embedding_dim],
                ),
            ],
            "ensemble_scheduling": dict(
                step=[
                    dict(
                        model_name="sentence_tokenizer",
                        model_version=-1,
                        input_map=dict(text="TEXT"),
                        output_map=dict(
                            input_ids="INPUT_IDS",
                            attention_mask="ATTENTION_MASK",
                        ),
                    ),
                    dict(
                        model_name="sentence_embeddings",
                        model_version=-1,
                        input_map=dict(
                            input_ids="INPUT_IDS",
                            attention_mask="ATTENTION_MASK",
                        ),
                        output_map=dict(embeddings="EMBEDDINGS"),
                    ),
                ]
            ),
        }

    def export_ensemble(self, triton_repo, triton_name, version=1):
        model_version_dir = init_model_repo(triton_repo, triton_name, version)

        config = self._create_ensemble_config()
        write_file = model_version_dir.parent / "config.pbtxt"
        save_triton_config(config, write_file)

    def export_tokenizer(self, triton_repo, triton_name, version=1):
        model_version_dir = init_model_repo(triton_repo, triton_name, version)

        shutil.copy(
            Path(__file__).parents[1]
            / "utils"
            / "triton_python"
            / "transformers_tokenizer.py",
            model_version_dir / "model.py",
        )

        config = self._create_tokenizer_config()
        write_config = model_version_dir.parent / "config.pbtxt"
        save_triton_config(config, write_config)

    @staticmethod
    def _create_tokenizer_config():
        return {
            "backend": "python",
            "max_batch_size": 16,
            "input": [dict(name="text", data_type="TYPE_STRING", dims=[-1])],
            "output": [
                dict(name="input_ids", data_type="TYPE_INT64", dims=[-1]),
                dict(name="attention_mask", data_type="TYPE_INT64", dims=[-1]),
            ],
            "parameters": {
                "EXECUTION_ENV_PATH": {
                    "string_value": "/opt/tritonserver/backends/python/py39.tar.gz"
                }
            },
        }



"""Run fiftyone.brain operations on a dataset.
"""
import fiftyone.brain as fob

from finegrained.utils.dataset import load_fiftyone_dataset


def compute_mistakenness(
    dataset: str, predictions: str, gt_field: str = "ground_truth", **kwargs
):
    """Estimate a probability that a ground truth label is wrong

    Args:
        dataset: fiftyone dataset name
        predictions: a field that contains model predictions
        gt_field: a field that contains ground truth data
        **kwargs: dataset loading filters

    Returns:
        none
    """
    dataset = load_fiftyone_dataset(dataset, **kwargs)
    fob.compute_mistakenness(dataset, predictions, label_field=gt_field)


def compute_hardness(dataset: str, predictions: str, **kwargs):
    """Estimate how difficult is this sample to predict.

    Args:
        dataset: fiftyone dataset name
        predictions: field with predictions
        **kwargs: dataset filters

    Returns:
        None
    """
    dataset = load_fiftyone_dataset(dataset, **kwargs)
    fob.compute_hardness(dataset, predictions)


def compute_uniqueness(dataset: str, label_field: str = "uniqueness", **kwargs):
    dataset = load_fiftyone_dataset(dataset, **kwargs)
    fob.compute_uniqueness(dataset, uniqueness_field=label_field)
    return dataset



"""Base constructs to use torchvision models.
"""
import fiftyone as fo
import torch
from torchvision.io.image import read_image
from torchvision.models.detection import (
    MaskRCNN_ResNet50_FPN_V2_Weights,
    maskrcnn_resnet50_fpn_v2,
)
from torchvision.transforms import functional as T
from tqdm import tqdm

from finegrained.models.torch_utils import get_device
from finegrained.utils.dataset import load_fiftyone_dataset


def _parse_bbox(box, image_size):
    h, w = image_size
    if h > w:
        box = box.T
    xmin, ymin, xmax, ymax = box.cpu() / torch.tensor([w, h, w, h])
    box_w = xmax - xmin
    box_h = ymax - ymin
    return [xmin.item(), ymin.item(), box_w.item(), box_h.item()]


def _parse_torchvision_detections(prediction, class_labels, image_size):
    detections = []
    for box, label, score in zip(
        prediction["boxes"], prediction["labels"], prediction["scores"]
    ):
        box = _parse_bbox(box, image_size)
        det = fo.Detection(
            label=class_labels[label.item()],
            bounding_box=box,
            condfidence=score.item(),
        )
        detections.append(det)
    return fo.Detections(detections=detections)


def _resize_image(img: torch.Tensor, target_size: int) -> torch.Tensor:
    orig_h, orig_w = img.shape[1:]
    max_dim = max(orig_h, orig_w)
    resize_h = int(orig_h * (target_size / max_dim))
    resize_w = int(orig_w * (target_size / max_dim))
    img = T.resize(img, size=(resize_h, resize_w))
    return img


def object_detection(
    dataset: str,
    label_field: str,
    conf: float = 0.25,
    image_size=None,
    device=None,
    **kwargs
):
    """Detect COCO objects with mask-rcnn-v2 from torchvision

    Args:
        dataset: fiftyone dataset name
        label_field: which field to write predictions to
        conf: box confidence threshold
        image_size: if specified, this will be a max image size
            (to save memory)
        **kwargs: dataset loading filters

    Returns:
        None
    """
    # prepare the model
    weights = MaskRCNN_ResNet50_FPN_V2_Weights.DEFAULT
    model = maskrcnn_resnet50_fpn_v2(weights=weights, box_score_thresh=conf)
    device = get_device()[0] if device is None else torch.device(device)
    model.eval().to(device)
    preprocess = weights.transforms()

    dataset = load_fiftyone_dataset(dataset, **kwargs)

    with torch.no_grad():
        for smp in tqdm(dataset.select_fields("filepath"), desc="detecting"):
            # prepare image input
            img = read_image(smp.filepath)
            if image_size:
                img = _resize_image(img, target_size=image_size)
            # predict
            batch = [preprocess(img).to(device)]
            prediction, *_ = model(batch)
            # parse detection and save results
            img_h, img_w = img.size(1), img.size(2)
            detections = _parse_torchvision_detections(
                prediction, weights.meta["categories"], (img_h, img_w)
            )
            smp[label_field] = detections
            smp.save()
